# /test/test_unused_warnings.py
#
# Test cases for style/* checks
#
# Disable no-self-use in this module as all test methods must be member
# functions, regardless of whether self is used.
#
# See /LICENCE.md for Copyright information
"""Test cases for style/* checks."""

from test import autoderef_list

from test.warnings_test_common import DEFINITION_TYPES
from test.warnings_test_common import FUNCTIONS_SETTING_VARS
from test.warnings_test_common import LinterFailure
from test.warnings_test_common import format_with_command
from test.warnings_test_common import gen_source_line
from test.warnings_test_common import run_linter_throw

from nose_parameterized import param, parameterized

from testtools import (ExpectedException, TestCase)


class TestPrivateFunctionsMustBeUsed(TestCase):
    """Test that private functions must be used."""

    @parameterized.expand(DEFINITION_TYPES)
    def test_pass_priv_func_used(self, definition):
        """unused/private passes if private function used."""
        script = ("{0} (_definition ARGUMENT)\n"
                  "end{0} ()\n"
                  "_definition (ARGUMENT)\n").format(definition)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/private"]))

    @parameterized.expand(DEFINITION_TYPES)
    def test_pass_priv_func_as_var(self, definition):
        """unused/private passes if private function used as a variable."""
        script = ("{0} (_definition ARGUMENT)\n"
                  "end{0} ()\n"
                  "call (_definition)\n").format(definition)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/private"]))

    @parameterized.expand(DEFINITION_TYPES)
    def test_pass_pub_func_unused(self, definition):
        """unused/private passes if public function unused."""
        script = ("{0} (definition ARGUMENT)\n"
                  "end{0} ()\n").format(definition)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/private"]))

    @parameterized.expand(DEFINITION_TYPES)
    def test_fail_priv_func_unused(self, definition):  # suppress(no-self-use)
        """unused/private passes if private function unused."""
        script = ("{0} (_definition ARGUMENT)\n"
                  "end{0} ()\n").format(definition)
        with ExpectedException(LinterFailure):
            run_linter_throw(script, whitelist=["unused/private"])

_FORMAT_WITH_DEREFFED_VAR = format_with_command(lambda x: "${" + x + "}")


def _format_with_generator(func, _, params):
    """Return formatted docstring with generated command."""
    cmd = params.args[0]
    generator = params.args[1]
    return func.__doc__.format("{0} ({1})".format(cmd,
                                                  generator(lambda x: ("_" +
                                                                       x))))


class TestUnusedSetVariablesInBody(TestCase):
    """Check for unused set variables in a function/macro body."""

    parameters = [param(m) for m in FUNCTIONS_SETTING_VARS]

    @parameterized.expand(parameters, testcase_func_doc=format_with_command())
    def test_pass_variable_used(self, matcher):
        """Check unused/var_in_func passes when var is used with {}."""
        script = ("function (f)\n"
                  "    {0}\n"
                  "    message ({1})\n"
                  "endfunction ()\n").format(gen_source_line(matcher),
                                             "${VALUE}")
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/var_in_func"]))

    @parameterized.expand(parameters, testcase_func_doc=format_with_command())
    def test_pass_nested_use(self, matcher):
        """Check unused/var_in_func passes with {}, with use nested."""
        script = ("function (f)\n"
                  "    {0}\n"
                  "    foreach (VAR LIST)\n"
                  "        message ({1} VAR)\n"
                  "    endforeach ()\n"
                  "endfunction ()\n").format(gen_source_line(matcher),
                                             "${VALUE}")
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/var_in_func"]))

    def test_global_used(self):
        """Check unused/var_in_func passes when global property is used."""
        script = ("function (f)\n"
                  "    set_property (GLOBAL PROPERTY VALUE)\n"
                  "endfunction (f)\n"
                  "function (g)\n"
                  "    get_property (VAR GLOBAL PROPERTY VALUE)\n"
                  "    message (${VAR})\n"
                  "endfunction ()\n")
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/var_in_func"]))

    @parameterized.expand(parameters,
                          testcase_func_doc=_FORMAT_WITH_DEREFFED_VAR)
    def test_pass_deref_unused(self, matcher):
        """Check unused/var_in_func passes when deref var is set with {}."""
        call = gen_source_line(matcher,
                               match_transform=lambda x: "${" + x + "}")
        script = ("function (f)\n"
                  "    {0}\n"
                  "endfunction ()\n").format(call)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/var_in_func"]))

    @parameterized.expand(autoderef_list.VARIABLES,
                          testcase_func_doc=_format_with_generator)
    def test_pass_use_autoderef(self, cmd, generator):
        """Check that unused/var_in_func passes when var autodereffed in {}."""
        script = ("function (f)\n"
                  "    set (_ARGUMENT 0)\n"
                  "    {0} ({1})\n"
                  "    end{0} ()\n"
                  "endfunction ()").format(cmd, generator(lambda x: "_" + x))
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/private_var"]))

    @parameterized.expand(parameters,
                          testcase_func_doc=_FORMAT_WITH_DEREFFED_VAR)
    def test_pass_compound_unused(self, matcher):
        """Check unused/var_in_func passes if compound_lit var passed in {}."""
        call = gen_source_line(matcher,
                               match_transform=lambda x: "${" + x + "}/Other")
        script = ("function (f)\n"
                  "    {0}\n"
                  "endfunction ()\n").format(call)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/var_in_func"]))

    @parameterized.expand(parameters, testcase_func_doc=format_with_command())
    def test_fail_variable_unused(self, matcher):  # suppress(no-self-use)
        """Check unused/var_in_func fails when var is unused in {}."""
        call = gen_source_line(matcher)
        script = ("function (f)\n"
                  "    {0}\n"
                  "endfunction ()\n").format(call)

        with ExpectedException(LinterFailure):
            run_linter_throw(script, whitelist=["unused/var_in_func"])

    @parameterized.expand(parameters, testcase_func_doc=format_with_command())
    def test_fail_nested_var_unused(self, matcher):  # suppress(no-self-use)
        """Check unused/var_in_func fails when nested var is unused in {}."""
        call = gen_source_line(matcher)
        script = ("function (f)\n"
                  "    if (COND)\n"
                  "        {0}\n"
                  "    endif (COND)\n"
                  "endfunction ()\n").format(call)

        with ExpectedException(LinterFailure):
            run_linter_throw(script, whitelist=["unused/var_in_func"])

# suppress(unnecessary-lambda)
_PRIVATE_VAR_SET_FORMAT = format_with_command(lambda x: "_{}".format(x))


class TestUnusedPrivateToplevelVars(TestCase):
    """Check for unused set private variables at the top level."""

    parameters = [param(m) for m in FUNCTIONS_SETTING_VARS]

    @parameterized.expand(parameters,
                          testcase_func_doc=_PRIVATE_VAR_SET_FORMAT)
    def test_pass_variable_used(self, matcher):
        """Check unused/private_var passes when var set by {} is used."""
        find = matcher.find
        xform = lambda x: "_{0}".format(x)  # suppress(unnecessary-lambda,E731)
        script = ("function (f)\n"
                  "    {0} ({1})\n"
                  "    message ({2})\n"
                  "endfunction ()\n").format(matcher.cmd,
                                             find.generate(matcher.sub,
                                                           lambda x: x,
                                                           xform),
                                             "${_VALUE}")
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/private_var"]))

    @parameterized.expand(parameters,
                          testcase_func_doc=_PRIVATE_VAR_SET_FORMAT)
    def test_pass_nested_use(self, matcher):
        """Check unused/private_var passes when {} is used in nested ctx."""
        find = matcher.find
        xform = lambda x: "_{0}".format(x)  # suppress(unnecessary-lambda,E731)
        script = ("{0} ({1})\n"
                  "function (f)\n"
                  "    foreach (VAR LIST)\n"
                  "        message ({2})\n"
                  "    endforeach ()\n"
                  "endfunction (f)\n").format(matcher.cmd,
                                              find.generate(matcher.sub,
                                                            lambda x: x,
                                                            xform),
                                              "${_VALUE}")
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/private_var"]))

    def test_global_priv_used(self):
        """Check unused/private_var passes when var is used in nested ctx."""
        script = ("function (f)\n"
                  "    set_property (GLOBAL PROPERTY _VALUE)\n"
                  "endfunction (f)\n"
                  "function (g)\n"
                  "    get_property (VAR GLOBAL PROPERTY _VALUE)\n"
                  "endfunction ()\n")
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/private_var"]))

    @parameterized.expand(parameters,
                          testcase_func_doc=_PRIVATE_VAR_SET_FORMAT)
    def test_pass_pub_var_unused(self, matcher):
        """Check unused/private_var passes when public var is unused in {}."""
        find = matcher.find
        script = ("{0} ({1})\n").format(matcher.cmd,
                                        find.generate(matcher.sub,
                                                      lambda x: x,
                                                      lambda x: x))
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/private_var"]))

    @parameterized.expand(autoderef_list.VARIABLES,
                          testcase_func_doc=_format_with_generator)
    def test_pass_use_var_autoderef(self, cmd, generator):
        """Check that unused/private_var passes when var autodereffed in {}."""
        script = ("set (_ARGUMENT 0)\n"
                  "{0} ({1})\n"
                  "end{0} ()\n").format(cmd, generator(lambda x: "_" + x))
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["unused/private_var"]))

    @parameterized.expand(parameters,
                          testcase_func_doc=_PRIVATE_VAR_SET_FORMAT)
    def test_fail_variable_unused(self, matcher):  # suppress(no-self-use)
        """Check unused/var_in_func fails when private var is unused in {}."""
        find = matcher.find
        xform = lambda x: "_{0}".format(x)  # suppress(unnecessary-lambda,E731)
        script = ("{0} ({1})\n").format(matcher.cmd,
                                        find.generate(matcher.sub,
                                                      lambda x: x,
                                                      xform))

        with ExpectedException(LinterFailure):
            run_linter_throw(script, whitelist=["unused/private_var"])



# /test/test_warnings.py
#
# Test cases for each warning in polysquare-cmake-linter
#
# See /LICENCE.md for Copyright information
"""Test the linter to ensure that each lint use-case triggers warnings."""

from polysquarecmakelinter import linter

from testtools import TestCase


class TestIgnoreLines(TestCase):
    """Test case for line-ignore logic."""

    def test_ignore_nolint_wildcard(self):
        """Ignore all lines with a wildcard NOLINT."""
        self.assertTrue(linter.should_ignore(" # NOLINT:*\n", "some/warning"))

    def test_no_ignore_unmatched_nolint(self):
        """Don't ignore lines that don't match the NOLINT."""
        self.assertFalse(linter.should_ignore(" # NOLINT:o/warning\n",
                                              "some/warning"))

    def test_no_ignore_malformed(self):
        """Don't ignore lines where NOLINT is malformed."""
        self.assertFalse(linter.should_ignore(" # NOLINT:\n",
                                              "some/warning"))

    def test_ignore_nolint_matching(self):
        """Ignore lines that match the NOLINT."""
        self.assertTrue(linter.should_ignore(" # NOLINT:some/warning\n",
                                             "some/warning"))



# /polysquarecmakelinter/find_set_variables.py
#
# Details to find set variables in a module
#
# See /LICENCE.md for Copyright information
"""Detail to find set variables in a module."""

from collections import namedtuple

from cmakeast import ast_visitor

from polysquarecmakelinter import ignore


# Subclass namedtuple so that we can have a default argument for sub
#
# suppress(too-few-public-methods)
class _SetVariable(namedtuple("_SetVariable",
                              "cmd find sub")):
    """_SetVariable namedtuple with default variable for sub."""

    def __new__(cls, cmd, find, sub=None):
        """Factory with sub defaulted out."""
        return super(_SetVariable, cls).__new__(cls, cmd, find, sub)


def _find_arg_no_factory(mixin):
    """Generate a function which returns a class with a mixin."""
    def _find_arg_no(num):
        """Return an argument finder for argument at num."""
        class Finder(mixin, object):  # suppress(too-few-public-methods)
            """Finder for arguments after num."""

            def __call__(self, arguments):  # suppress(no-self-use)
                """Return argument at num."""
                try:
                    return arguments[num]
                except IndexError:
                    return None

        return Finder()

    return _find_arg_no


def _find_after_arg_factory(mixin):
    """Generate a function which returns a class with a mixin."""
    def _find_after_arg(argument):
        """Return an argument finder for an argument after argument."""
        class Finder(mixin, object):  # suppress(too-few-public-methods)
            """Finder for arguments after argument."""

            def __call__(self, arguments):  # suppress(no-self-use)
                """Return argument after argument or None if not found."""
                try:
                    argument_index = 1
                    for argument_index in range(1, len(arguments)):
                        if arguments[argument_index - 1].contents == argument:
                            break

                    return arguments[argument_index]
                except IndexError:
                    return None

        return Finder()

    return _find_after_arg


class _EmptyMixin(object):  # suppress(too-few-public-methods)
    """An empty mixin to pass to argument factories."""


def all_functions(arg_no_mixin_generator=lambda n: _EmptyMixin,
                  after_arg_mixin_generator=lambda a: _EmptyMixin):
    """Return a list of _SetVariable for all functions setting variables.

    Use arg_no_mixin_generator and after_arg_mixin_generator to customize the
    instance of the find attribute attached to each _SetVariable, eg, to add
    your own functions.
    """
    def _find_arg_no(num):
        """Wrap _find_arg_no_factory and adds mixin."""
        return _find_arg_no_factory(arg_no_mixin_generator(num))(num)

    def _find_after_arg(arg):
        """Wrap _find_after_arg_factory and adds mixin."""
        return _find_after_arg_factory(after_arg_mixin_generator(arg))(arg)

    return [
        _SetVariable("aux_source_directory", _find_arg_no(1)),
        _SetVariable("build_command", _find_arg_no(0)),
        _SetVariable("cmake_host_system_information", _find_arg_no(1),
                     sub="RESULT"),
        _SetVariable("cmake_policy", _find_arg_no(2), sub="GET"),
        _SetVariable("execute_process", _find_after_arg("RESULT_VARIABLE")),
        _SetVariable("execute_process", _find_after_arg("OUTPUT_VARIABLE")),
        _SetVariable("execute_process", _find_after_arg("ERROR_VARIABLE")),
        _SetVariable("file", _find_arg_no(2), sub="READ"),
        _SetVariable("file", _find_arg_no(2), sub="STRINGS"),
        _SetVariable("file", _find_arg_no(2), sub="MD5"),
        _SetVariable("file", _find_arg_no(2), sub="SHA1"),
        _SetVariable("file", _find_arg_no(2), sub="SHA224"),
        _SetVariable("file", _find_arg_no(2), sub="SHA256"),
        _SetVariable("file", _find_arg_no(2), sub="SHA384"),
        _SetVariable("file", _find_arg_no(2), sub="SHA512"),
        _SetVariable("file", _find_arg_no(1), sub="GLOB"),
        _SetVariable("file", _find_arg_no(1), sub="GLOB_RECURSE"),
        _SetVariable("file", _find_arg_no(1), sub="RELATIVE_PATH"),
        _SetVariable("file", _find_arg_no(2), sub="TO_CMAKE_PATH"),
        _SetVariable("file", _find_arg_no(2), sub="TO_NATIVE_PATH"),
        _SetVariable("file", _find_after_arg("LOG"), sub="DOWNLOAD"),
        _SetVariable("file", _find_after_arg("LOG"), sub="UPLOAD"),
        _SetVariable("file", _find_after_arg("STATUS"), sub="DOWNLOAD"),
        _SetVariable("file", _find_after_arg("STATUS"), sub="UPLOAD"),
        _SetVariable("file", _find_arg_no(2), sub="TIMESTAMP"),
        _SetVariable("find_file", _find_arg_no(0)),
        _SetVariable("find_library", _find_arg_no(0)),
        _SetVariable("find_path", _find_arg_no(0)),
        _SetVariable("find_program", _find_arg_no(0)),
        _SetVariable("get_cmake_property", _find_arg_no(0)),
        _SetVariable("get_directory_property", _find_arg_no(0)),
        _SetVariable("get_filename_component", _find_arg_no(0)),
        _SetVariable("get_property", _find_arg_no(0)),
        _SetVariable("get_source_file_property", _find_arg_no(1)),
        _SetVariable("get_target_property", _find_arg_no(0)),
        _SetVariable("get_test_property", _find_arg_no(0)),
        _SetVariable("include", _find_after_arg("RESULT_VARIABLE")),
        _SetVariable("separate_arguments", _find_arg_no(1)),
        _SetVariable("set", _find_arg_no(0)),
        _SetVariable("unset", _find_arg_no(0)),
        _SetVariable("try_compile", _find_after_arg("RESULT_VAR")),
        _SetVariable("try_compile", _find_after_arg("OUTPUT_VARIABLE")),
        _SetVariable("try_run", _find_after_arg("RUN_RESULT_VAR")),
        _SetVariable("try_run", _find_after_arg("COMPILE_RESULT_VAR")),
        _SetVariable("try_run", _find_after_arg("RUN_OUTPUT_VARIABLE")),
        _SetVariable("try_run", _find_after_arg("OUTPUT_VARIABLE")),
        _SetVariable("list", _find_arg_no(2), sub="LENGTH"),
        _SetVariable("list", _find_arg_no(-1), sub="GET"),
        _SetVariable("list", _find_arg_no(3), sub="FIND"),
        _SetVariable("list", _find_arg_no(2), sub="REMOVE_ITEM"),
        _SetVariable("list", _find_arg_no(1), sub="REMOVE_AT"),
        _SetVariable("list", _find_arg_no(1), sub="REMOVE_DUPLICATES"),
        _SetVariable("list", _find_arg_no(1), sub="REVERSE"),
        _SetVariable("list", _find_arg_no(1), sub="SORT"),
        _SetVariable("list", _find_arg_no(1), sub="APPEND"),
        _SetVariable("list", _find_arg_no(1), sub="INSERT"),
        _SetVariable("match", _find_arg_no(1), sub="EXPR")
    ]

# Default list of _SetVariable functions.
_FIND_AFTER_ARG_INT = _find_after_arg_factory(_EmptyMixin)

FUNCTIONS_SETTING_VARIABLES = all_functions()
_FUNCTIONS_SETTING_VARS_INT = (FUNCTIONS_SETTING_VARIABLES +
                               [_SetVariable("set_property",
                                             _FIND_AFTER_ARG_INT("PROPERTY"),
                                             sub="GLOBAL")])


def by_function_call(node):
    """Return a variable (as a word node) set by a function call, if any."""
    for matcher in _FUNCTIONS_SETTING_VARS_INT:
        if matcher.cmd == node.name:
            # Exclude where subcommand does not match
            # Note: pylint trips up on the generated code and believes that
            # matcher has no attribute named "sub". This is incorrect, so
            # a warning is suppressed here.
            if matcher.sub is not None:  # suppress(no-member)
                try:
                    if (matcher.sub !=  # suppress(no-member)
                            node.arguments[0].contents):
                        continue
                except IndexError:
                    continue

            evaluate = matcher.find(node.arguments)

            if evaluate:
                return evaluate

    return None


def in_tree(abstract_syntax_tree):
    """Find all assigned variables in this abstract_syntax_tree."""
    variables = []

    def _call_visitor(name, node):
        """Visit all function calls."""
        assert name == "FunctionCall"

        evaluate = by_function_call(node)
        if evaluate:
            variables.append(evaluate)

    ast_visitor.recurse(abstract_syntax_tree,
                        function_call=ignore.visitor_depth(_call_visitor))

    return variables



# /test/autoderef_list.py
#
# Utility helpers to generate situations where variables might be automatically
# dereferenced.
#
# See /LICENCE.md for Copyright information
"""Generate situations where variables are automatically dereferenced."""

from collections import namedtuple

_AutoderefVariable = namedtuple("_AutoderefVariable", "cmd find")

KNOWN_KEYWORDS = {
    "if": [
        "NOT"
        "AND"
        "OR"
        "STREQUAL"
        "STRLESS"
        "STRGREATER"
        "LESS"
        "GREATER"
        "EQUAL"
        "POLICY"
        "TARGET"
        "EXISTS"
        "IS_NEWER_THAN"
        "DEFINED"
        "VERSION_LESS",
        "VERSION_EQUAL",
        "VERSION_GREATER",
        "ON",
        "OFF",
        "TRUE",
        "FALSE"
    ],
    "foreach": [
        "RANGE",
        "IN"
        "LISTS",
        "ITEMS"
    ]
}

KNOWN_KEYWORDS["while"] = KNOWN_KEYWORDS["if"]
KNOWN_KEYWORDS["elseif"] = KNOWN_KEYWORDS["if"]


def _first_arg():
    """Return a finder for first argument."""
    def generate(transform):
        """Generate one arg."""
        return transform("ARGUMENT")

    return generate


def _last_arg():
    """Return a finder for last argument."""
    def generate(transform):
        """Generate two args, last one being ARGUMENT."""
        return "ONE {0}".format(transform("ARGUMENT"))

    return generate


def _after_arg(arg_name):
    """Return a finder for argument after arg_name."""
    def generate(transform):
        """Generate three args, OTHER, arg_name and ARGUMENT."""
        return "ONE {0} {1}".format(arg_name, transform("ARGUMENT"))

    return generate


def _before_arg(arg_name):
    """Return a finder for argument before arg_name."""
    def generate(transform):
        """Generate three args, ARGUMENT, arg_name and OTHER."""
        return "ONE {0} {1}".format(transform("ARGUMENT"), arg_name)

    return generate


def _handle_autoderef(cmd, generator):
    """Wrap a finder to handle argument ignoring."""
    return (cmd, generator)

# This constant is used in the unused warnings test.
#
# suppress(unused-variable)
VARIABLES = [
    _handle_autoderef("if", _first_arg()),
    _handle_autoderef("if", _after_arg("NOT")),
    _handle_autoderef("if", _before_arg("AND")),
    _handle_autoderef("if", _after_arg("AND")),
    _handle_autoderef("if", _before_arg("OR")),
    _handle_autoderef("if", _after_arg("OR")),
    _handle_autoderef("if", _before_arg("OR")),
    _handle_autoderef("if", _after_arg("OR")),
    _handle_autoderef("if", _before_arg("STREQUAL")),
    _handle_autoderef("if", _after_arg("STREQUAL")),
    _handle_autoderef("if", _before_arg("STRLESS")),
    _handle_autoderef("if", _after_arg("STRLESS")),
    _handle_autoderef("if", _before_arg("STRGREATER")),
    _handle_autoderef("if", _after_arg("STRGREATER")),
    _handle_autoderef("if", _before_arg("LESS")),
    _handle_autoderef("if", _after_arg("LESS")),
    _handle_autoderef("if", _before_arg("GREATER")),
    _handle_autoderef("if", _after_arg("GREATER")),
    _handle_autoderef("if", _before_arg("EQUAL")),
    _handle_autoderef("if", _after_arg("EQUAL")),
    _handle_autoderef("if", _before_arg("VERSION_LESS")),
    _handle_autoderef("if", _after_arg("VERSION_LESS")),
    _handle_autoderef("if", _before_arg("VERSION_GREATER")),
    _handle_autoderef("if", _after_arg("VERSION_GREATER")),
    _handle_autoderef("if", _before_arg("VERSION_EQUAL")),
    _handle_autoderef("if", _after_arg("VERSION_EQUAL")),
    _handle_autoderef("if", _before_arg("MATCHES")),
    _handle_autoderef("if", _after_arg("DEFINED")),
    _handle_autoderef("while", _first_arg()),
    _handle_autoderef("while", _after_arg("NOT")),
    _handle_autoderef("while", _before_arg("AND")),
    _handle_autoderef("while", _after_arg("AND")),
    _handle_autoderef("while", _before_arg("OR")),
    _handle_autoderef("while", _after_arg("OR")),
    _handle_autoderef("while", _before_arg("OR")),
    _handle_autoderef("while", _after_arg("OR")),
    _handle_autoderef("while", _before_arg("STREQUAL")),
    _handle_autoderef("while", _after_arg("STREQUAL")),
    _handle_autoderef("while", _before_arg("STRLESS")),
    _handle_autoderef("while", _after_arg("STRLESS")),
    _handle_autoderef("while", _before_arg("STRGREATER")),
    _handle_autoderef("while", _after_arg("STRGREATER")),
    _handle_autoderef("while", _before_arg("LESS")),
    _handle_autoderef("while", _after_arg("LESS")),
    _handle_autoderef("while", _before_arg("GREATER")),
    _handle_autoderef("while", _after_arg("GREATER")),
    _handle_autoderef("while", _before_arg("EQUAL")),
    _handle_autoderef("while", _after_arg("EQUAL")),
    _handle_autoderef("while", _before_arg("VERSION_LESS")),
    _handle_autoderef("while", _after_arg("VERSION_LESS")),
    _handle_autoderef("while", _before_arg("VERSION_GREATER")),
    _handle_autoderef("while", _after_arg("VERSION_GREATER")),
    _handle_autoderef("while", _before_arg("VERSION_EQUAL")),
    _handle_autoderef("while", _after_arg("VERSION_EQUAL")),
    _handle_autoderef("while", _before_arg("MATCHES")),
    _handle_autoderef("while", _after_arg("DEFINED")),
    _handle_autoderef("elseif", _first_arg()),
    _handle_autoderef("elseif", _after_arg("NOT")),
    _handle_autoderef("elseif", _before_arg("AND")),
    _handle_autoderef("elseif", _after_arg("AND")),
    _handle_autoderef("elseif", _before_arg("OR")),
    _handle_autoderef("elseif", _after_arg("OR")),
    _handle_autoderef("elseif", _before_arg("OR")),
    _handle_autoderef("elseif", _after_arg("OR")),
    _handle_autoderef("elseif", _before_arg("STREQUAL")),
    _handle_autoderef("elseif", _after_arg("STREQUAL")),
    _handle_autoderef("elseif", _before_arg("STRLESS")),
    _handle_autoderef("elseif", _after_arg("STRLESS")),
    _handle_autoderef("elseif", _before_arg("STRGREATER")),
    _handle_autoderef("elseif", _after_arg("STRGREATER")),
    _handle_autoderef("elseif", _before_arg("LESS")),
    _handle_autoderef("elseif", _after_arg("LESS")),
    _handle_autoderef("elseif", _before_arg("GREATER")),
    _handle_autoderef("elseif", _after_arg("GREATER")),
    _handle_autoderef("elseif", _before_arg("EQUAL")),
    _handle_autoderef("elseif", _after_arg("EQUAL")),
    _handle_autoderef("elseif", _before_arg("VERSION_LESS")),
    _handle_autoderef("elseif", _after_arg("VERSION_LESS")),
    _handle_autoderef("elseif", _before_arg("VERSION_GREATER")),
    _handle_autoderef("elseif", _after_arg("VERSION_GREATER")),
    _handle_autoderef("elseif", _before_arg("VERSION_EQUAL")),
    _handle_autoderef("elseif", _after_arg("VERSION_EQUAL")),
    _handle_autoderef("elseif", _before_arg("MATCHES")),
    _handle_autoderef("elseif", _after_arg("DEFINED")),
    _handle_autoderef("foreach", _last_arg()),
    _handle_autoderef("list", _after_arg("LENGTH")),
    _handle_autoderef("list", _after_arg("GET")),
    _handle_autoderef("list", _after_arg("FIND"))
]



# /polysquarecmakelinter/ignore.py
#
# Wrapper functions to ignore certain arguments in callbacks
#
# See /LICENCE.md for Copyright information
"""Wrapper functions to ignore certain arguments in callbacks."""


def all_but_ast(check):
    """Only passes AST to check."""
    def _check_wrapper(contents, ast, **kwargs):
        """Wrap check and passes the AST to it."""
        del contents
        del kwargs

        return check(ast)

    return _check_wrapper


def check_kwargs(check):
    """Return wrapper for check function."""
    def _check_wrapper(contents, ast, **kwargs):
        """Do not pass kwargs to check."""
        del kwargs

        return check(contents, ast)

    return _check_wrapper


def visitor_depth(visitor):
    """Return wrapper for depth function."""
    def _visitor_wrapper(name, node, depth):
        """Do not pass depth to visitor."""
        del depth

        visitor(name, node)

    return _visitor_wrapper



# /test/test_find_variable_scopes.py
#
# Tests that we're able to find variables inside the scopes that we expect
#
# See /LICENCE.md for Copyright information
"""Test that we're able to find variables inside the scopes that we expect."""

from test.warnings_test_common import Equals
from test.warnings_test_common import FUNCTIONS_SETTING_VARS
from test.warnings_test_common import format_with_command
from test.warnings_test_common import gen_source_line

from cmakeast import ast

from nose_parameterized import param, parameterized

from polysquarecmakelinter import find_variables_in_scopes

from polysquarecmakelinter.find_variables_in_scopes import VariableSource

from testtools import TestCase

from testtools.matchers import MatchesStructure
from testtools.matchers import Not

VarSrc = VariableSource


class TestFindVariablesInScopes(TestCase):
    """Test fixture for the in_tree function."""

    params = [param(m) for m in FUNCTIONS_SETTING_VARS]

    @parameterized.expand(params, testcase_func_doc=format_with_command())
    def test_global_scope(self, matcher):
        """Test setting and finding vars with {} at global scope."""
        script = "{0}".format(gen_source_line(matcher))
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))

        self.assertThat(global_scope.set_vars[0].node,
                        MatchesStructure(contents=Equals("VALUE")))

    @parameterized.expand(params, testcase_func_doc=format_with_command())
    def test_in_func_scope(self, matcher):
        """Test setting and finding vars with {} in function scope."""
        script = ("function (foo)\n"
                  "    {0}\n"
                  "endfunction ()\n").format(gen_source_line(matcher))
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))

        self.assertThat(global_scope.scopes[0].set_vars[0].node,
                        MatchesStructure(contents=Equals("VALUE")))

    @parameterized.expand(params, testcase_func_doc=format_with_command())
    def test_if_in_func_scope(self, matcher):
        """Test that using {} in an if block propagates variable to func."""
        script = ("function (foo)\n"
                  "    if (CONDITION)\n"
                  "        {0}\n"
                  "    endif (CONDITION)\n"
                  "endfunction ()\n").format(gen_source_line(matcher))
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].set_vars[0].node,
                        MatchesStructure(contents=Equals("VALUE")))

    @parameterized.expand(params, testcase_func_doc=format_with_command())
    def test_elseif_in_func_scope(self, matcher):
        """Test that using {} in an elseif statement propagates variable."""
        script = ("function (foo)\n"
                  "    if (CONDITION)\n"
                  "    elseif (CONDITION)\n"
                  "        {0}\n"
                  "    endif (CONDITION)\n"
                  "endfunction ()\n").format(gen_source_line(matcher))
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].set_vars[0].node,
                        MatchesStructure(contents=Equals("VALUE")))

    @parameterized.expand(params, testcase_func_doc=format_with_command())
    def test_else_in_func_scope(self, matcher):
        """Test that using {} in an else statement propagates variable."""
        script = ("function (foo)\n"
                  "    if (CONDITION)\n"
                  "    else (CONDITION)\n"
                  "        {0}\n"
                  "    endif (CONDITION)\n"
                  "endfunction ()\n").format(gen_source_line(matcher))
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].set_vars[0].node,
                        MatchesStructure(contents=Equals("VALUE")))

    @parameterized.expand(params, testcase_func_doc=format_with_command())
    def test_while_in_func(self, matcher):
        """Test that using {} in a while block propagates variable to func."""
        script = ("function (foo)\n"
                  "    while (CONDITION)\n"
                  "        {0}\n"
                  "    endwhile (CONDITION)\n"
                  "endfunction ()\n").format(gen_source_line(matcher))
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].set_vars[0].node,
                        MatchesStructure(contents=Equals("VALUE")))

    @parameterized.expand(params, testcase_func_doc=format_with_command())
    def test_foreach_in_func(self, matcher):
        """Test that using {} in an foreach statements propagates variable."""
        script = ("function (foo)\n"
                  "    foreach (VAR LISTVAR)\n"
                  "        {0}\n"
                  "    endforeach ()\n"
                  "endfunction ()\n").format(gen_source_line(matcher))
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].set_vars[0].node,
                        MatchesStructure(contents=Equals("VALUE")))

    def test_foreach_scope(self):
        """Test that setting a variable in an foreach statements propagates."""
        script = ("function (foo)\n"
                  "    foreach (VAR LISTVAR)\n"
                  "    endforeach ()\n"
                  "endfunction ()\n")
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].scopes[0].set_vars[0].node,
                        MatchesStructure(contents=Equals("VAR")))

        foreach_type = find_variables_in_scopes.ScopeType.Foreach
        self.assertThat(global_scope.scopes[0].scopes[0].info,
                        MatchesStructure(type=Equals(foreach_type)))

    def test_parent_scope(self):
        """Test that setting a variable in the parent scope propagates."""
        script = ("function (foo)\n"
                  "    function (other)\n"
                  "        set (VARIABLE OTHER PARENT_SCOPE)\n"
                  "    endfunction ()\n"
                  "endfunction ()\n")
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].set_vars[0].node,
                        MatchesStructure(contents=Equals("VARIABLE")))

    def test_cache_scope(self):
        """Test that setting a variable in the cache scope is global."""
        script = ("function (foo)\n"
                  "    function (other)\n"
                  "        set (VARIABLE OTHER CACHE)\n"
                  "    endfunction ()\n"
                  "endfunction ()\n")
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.set_vars[0].node,
                        MatchesStructure(contents=Equals("VARIABLE")))

    def test_global_setprop_scope(self):
        """Test that setting a variable in the set_property scope is global."""
        script = ("function (foo)\n"
                  "    function (other)\n"
                  "        set_property (GLOBAL PROPERTY VARIABLE OTHER)\n"
                  "    endfunction ()\n"
                  "endfunction ()\n")
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.set_vars[0].node,
                        MatchesStructure(contents=Equals("VARIABLE")))

    def test_func_var_scope(self):
        """Test function variable scope."""
        script = ("function (foo VAR)\n"
                  "endfunction ()")
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].set_vars[0].node,
                        MatchesStructure(contents=Equals("VAR")))

    def test_macro_var_scope(self):
        """Test macro variable scope."""
        script = ("macro (foo VAR)\n"
                  "endmacro ()")
        global_scope = find_variables_in_scopes.set_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].set_vars[0].node,
                        MatchesStructure(contents=Equals("VAR")))

VARIABLE_USAGE_METHODS = [
    "${VALUE}",
    "${VALUE}_",
    "${VALUE}_${OTHER}",
    "${VALUE}/${OTHER}",
    "${VARIABLE_${VALUE}_OTHER}",
    "VALUE",
]


class TestUsedInTree(TestCase):
    """Test fixture for used_in_tree func."""

    @parameterized.expand(VARIABLE_USAGE_METHODS)
    def test_use_at_toplevel(self, call):
        """Test that a variable is marked as used at the toplevel."""
        script = "f ({0})".format(call)
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.used_vars[0],
                        MatchesStructure(source=Equals(VarSrc.GlobalVar)))

    @parameterized.expand(VARIABLE_USAGE_METHODS)
    def test_used_in_func(self, call):
        """Test that a variable is marked as used in a function."""
        script = ("function (name)\n"
                  "    f ({0})\n"
                  "endfunction ()\n").format(call)
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].used_vars[0],
                        MatchesStructure(source=Equals(VarSrc.FunctionVar)))

    @parameterized.expand(VARIABLE_USAGE_METHODS)
    def test_used_in_macro(self, call):
        """Test that a variable is marked as used in a macro."""
        script = ("macro (name)\n"
                  "    f ({0})\n"
                  "endmacro ()\n").format(call)
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].used_vars[0],
                        MatchesStructure(source=Equals(VarSrc.MacroVar)))

    @parameterized.expand(VARIABLE_USAGE_METHODS)
    def test_used_in_func_nest(self, call):
        """Test that a variable is marked as used in a function when nested."""
        script = ("function (name)\n"
                  "    if ()"
                  "        f ({0})\n"
                  "    endif ()"
                  "endfunction ()\n").format(call)
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].used_vars[0],
                        MatchesStructure(source=Equals(VarSrc.FunctionVar)))

    @parameterized.expand(VARIABLE_USAGE_METHODS)
    def test_used_in_func_if(self, call):
        """Test that a variable is marked as used in a function when nested."""
        script = ("function (name)\n"
                  "    if ({0})"
                  "    endif ()"
                  "endfunction ()\n").format(call)
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].used_vars[0],
                        MatchesStructure(source=Equals(VarSrc.FunctionVar)))

    @parameterized.expand(VARIABLE_USAGE_METHODS)
    def test_used_in_macro_if(self, call):
        """Test that a variable is marked as used in a function when nested."""
        script = ("macro (name)\n"
                  "    if ({0})"
                  "    endif ()"
                  "endmacro ()\n").format(call)
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].used_vars[0],
                        MatchesStructure(source=Equals(VarSrc.MacroVar)))

    @parameterized.expand(VARIABLE_USAGE_METHODS)
    def test_used_in_func_foreach(self, call):
        """Test that a variable is marked as used in a function when nested."""
        script = ("function (name)\n"
                  "    foreach (VAR {0})"
                  "    endforeach ()"
                  "endfunction ()\n").format(call)
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].used_vars[0],
                        MatchesStructure(source=Equals(VarSrc.FunctionVar)))

    @parameterized.expand(VARIABLE_USAGE_METHODS)
    def test_used_in_macro_foreach(self, call):
        """Test that a variable is marked as used in a function when nested."""
        script = ("macro (name)\n"
                  "    foreach (VAR {0})"
                  "    endforeach ()"
                  "endmacro ()\n").format(call)
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.scopes[0].used_vars[0],
                        MatchesStructure(source=Equals(VarSrc.MacroVar)))

    def test_not_used_in_function_hdr(self):
        """Test that there is no use in a function header."""
        script = ("function (name ARGUMENT)\n"
                  "endfunction ()")
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertEqual(len(global_scope.scopes[0].used_vars), 0)

    def test_no_use_by_foreach_var(self):
        """Test that there is no use for a foreach var."""
        script = ("foreach (VAR ${LIST})\n"
                  "endforeach ()")
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.used_vars[0].node,
                        MatchesStructure(contents=Not(Equals("VAR"))))

    @parameterized.expand(find_variables_in_scopes.FOREACH_KEYWORDS)
    def test_exclude_foreach_kws(self, keyword):
        """Test that there is no use for a foreach keyword."""
        script = ("foreach (VAR {0} LIST)\n"
                  "endforeach ()").format(keyword)
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.used_vars[0].node,
                        MatchesStructure(contents=Not(Equals(keyword))))

    @parameterized.expand(find_variables_in_scopes.IF_KEYWORDS)
    def test_exclude_if_kws(self, keyword):
        """Test that there is no use for an if keyword."""
        script = ("if ({0} OTHER)\n"
                  "endif ()").format(keyword)
        global_scope = find_variables_in_scopes.used_in_tree(ast.parse(script))
        self.assertThat(global_scope.used_vars[0].node,
                        MatchesStructure(contents=Not(Equals(keyword))))



# /polysquarecmakelinter/check_style.py
#
# Linter checks for style guide
#
# See /LICENCE.md for Copyright information
"""Linter checks for style guide."""

import re

from collections import namedtuple

from cmakeast import ast_visitor

from cmakeast.ast import WordType

from polysquarecmakelinter import find_set_variables
from polysquarecmakelinter import ignore
from polysquarecmakelinter import util

from polysquarecmakelinter.types import LinterFailure

_RE_DOUBLE_OUTER_QUOTES = re.compile("^\".*\"$")
_RE_IS_DEFINITION = re.compile(r"function|macro")

_RE_HEADERBODY_LIKE = re.compile(r"WhileStatement|"
                                 "ForeachStatement|"
                                 "FunctionDefinition|"
                                 "MacroDefinition")
_RE_FUNCTION_CALL = re.compile(r"FunctionCall")
_RE_IF_BLOCK = re.compile(r"IfBlock")
_RE_TOPLEVEL = re.compile(r"ToplevelBody")


def space_before_call(contents, abstract_syntax_tree):
    """Check that each function call is preceded by a single space."""
    errors = []

    def _call_handler(name, node):
        """Handle function calls."""
        assert name == "FunctionCall"

        line_index = node.line - 1
        col_index = node.col - 1
        end_of_name_index = col_index + len(node.name)
        num_spaces_until_openparen = 0
        index = end_of_name_index
        while contents[line_index][index] != "(":
            index += 1

        # AST checks will assert if the open paren is not found here

        num_spaces_until_openparen = index - end_of_name_index

        # Must be one space only
        if num_spaces_until_openparen != 1:
            extra_spaces = num_spaces_until_openparen - 1
            msg = "{0} extra spaces between call of {1}".format(extra_spaces,
                                                                node.name)
            replacement = util.replace_word(contents[line_index],
                                            end_of_name_index,
                                            " " * (extra_spaces + 1),
                                            " ")
            errors.append(LinterFailure(msg, node.line, replacement))

    ast_visitor.recurse(abstract_syntax_tree,
                        function_call=ignore.visitor_depth(_call_handler))

    return errors


def lowercase_functions(contents, abstract_syntax_tree):
    """Check that function / macro usage is all lowercase."""
    errors = []

    def _call_handler(name, node):
        """Handle function calls."""
        assert name == "FunctionCall"

        if node.name.lower() != node.name:
            msg = "{0} is not lowercase".format(node.name)
            replacement = util.replace_word(contents[node.line - 1],
                                            node.col - 1,
                                            node.name,
                                            node.name.lower())
            errors.append(LinterFailure(msg, node.line, replacement))

    def _definition_handler(name, node):
        """Handle function calls and macro definitions."""
        assert name == "FunctionDefinition" or name == "MacroDefinition"

        definition_name_word = node.header.arguments[0]
        definition_name = definition_name_word.contents

        if definition_name.lower() != definition_name:
            msg = "{0} name {1} is not lowercase".format(name, definition_name)
            line_index = definition_name_word.line - 1
            col_index = definition_name_word.col - 1
            replacement = util.replace_word(contents[line_index],
                                            col_index,
                                            definition_name,
                                            definition_name.lower())
            errors.append(LinterFailure(msg,
                                        definition_name_word.line,
                                        replacement))

    ast_visitor.recurse(abstract_syntax_tree,
                        function_call=ignore.visitor_depth(_call_handler),
                        function_def=ignore.visitor_depth(_definition_handler),
                        macro_def=ignore.visitor_depth(_definition_handler))

    return errors


def uppercase_arguments(contents, abstract_syntax_tree):
    """Check that arguments to definitions are all uppercase."""
    errors = []

    def _definition_visitor(name, node):
        """Visit all definitions."""
        assert name == "FunctionDefinition" or name == "MacroDefinition"

        if len(node.header.arguments) < 2:
            return

        for arg in node.header.arguments[1:]:
            if arg.contents.upper() != arg.contents:
                msg = "{0} must be uppercase".format(arg.contents)
                line_index = arg.line - 1
                col_index = arg.col - 1
                replacement = util.replace_word(contents[line_index],
                                                col_index,
                                                arg.contents,
                                                arg.contents.upper())
                errors.append(LinterFailure(msg, arg.line, replacement))

    ast_visitor.recurse(abstract_syntax_tree,
                        function_def=ignore.visitor_depth(_definition_visitor),
                        macro_def=ignore.visitor_depth(_definition_visitor))

    return errors


def set_variables_capitalized(contents, abstract_syntax_tree):
    """Check that each variable mutated is capitalized."""
    errors = []

    variables = find_set_variables.in_tree(abstract_syntax_tree)
    for evaluate in variables:

        evaluate_upper = evaluate.contents.upper()

        # Argument should be either String or Variable and a transformation
        # to uppercase should have no effect
        if (util.is_word_sink_variable(evaluate.type) and
                evaluate_upper != evaluate.contents):
            desc = "{0} must be uppercase".format(evaluate.contents)
            line = evaluate.line - 1
            replacement = util.replace_word(contents[line],
                                            evaluate.col - 1,
                                            evaluate.contents,
                                            evaluate_upper)
            errors.append(LinterFailure(desc,
                                        evaluate.line,
                                        replacement))
            break

    return errors

AlignmentInfo = namedtuple("AlignmentInfo", "col line line_has_kw")


def _expected_alignment(arg,
                        baseline_col,
                        baseline_line,
                        current_align):
    """Get expected alignment for the function argument arg."""
    col = current_align.col
    line = current_align.line
    line_has_kw = current_align.line_has_kw

    if arg.col == baseline_col:
        # Baseline column - reset align.col

        if (arg.type == WordType.Variable or
                arg.line == baseline_line):
            # This might be a keyword arg - allow alignment to this
            # line. Otherwise it is the first argument on this line,
            # or a function call where we allow arguments on any line
            line = arg.line
        else:
            # This is not a keyword arg. Do not allow alignment to
            # this line (only this column)
            line = None

        line_has_kw = (arg.type == WordType.Variable)
        col = arg.col

    # If we're on our currently aligned line, and this line starts
    # with a keyword-like argument, then allow align.col
    # to shift appropriately, check horizontal space
    if line == arg.line and line_has_kw:
        col = arg.col

    return AlignmentInfo(col, line, line_has_kw)


def _check_horizontal_space(node, index, contents):
    """Check horizontal space between arguments on same line."""
    if index > 0:
        current_column = node.arguments[index].col - 1
        previous_column = node.arguments[index - 1].col - 1
        previous_len = len(node.arguments[index - 1].contents)
        spaces_start = previous_column + previous_len
        num_spaces = current_column - (previous_column + previous_len)

        line_number = node.arguments[index].line - 1

        if num_spaces != 1:
            cur = node.arguments[index].contents
            prev = node.arguments[index - 1].contents

            msg = "Must be a single space between {0} and {1}".format(cur,
                                                                      prev)
            replacement = util.replace_word(contents[line_number],
                                            spaces_start,
                                            " " * num_spaces,
                                            " ")
            return LinterFailure(msg, node.line, replacement)


def _check_alignment(arg,
                     last_align,
                     baseline_col,
                     baseline_line,
                     contents):
    """Check alignment of arg."""
    align = _expected_alignment(arg,
                                baseline_col,
                                baseline_line,
                                last_align)

    misaligned_col = align.col and align.col != arg.col
    misaligned_line = align.line and align.line != arg.line

    def _offset_with_space(line, align_col, arg_col):
        """Insert spaces as required to place arg at align_col."""
        offset = max(0, arg_col - align_col)
        spaces = max(0, align_col - arg_col)
        return util.replace_word(line,
                                 arg_col - 1 - offset,
                                 " " * offset,
                                 " " * spaces)

    if ((misaligned_col or not align.col) and
            (misaligned_line or not align.line)):

        msg_parts_append_table = [
            (lambda: misaligned_line is not None,
             "line {0}".format(align.line)),
            (lambda: misaligned_col is not None,
             "col {0}".format(align.col)),
            (lambda: align.col != baseline_col,
             "col {0}".format(baseline_col))
        ]

        replacement = None
        msg_parts = [msg for c, msg in msg_parts_append_table if c()]

        if misaligned_col and align.col == baseline_col:
            replacement = _offset_with_space(contents[arg.line - 1],
                                             align.col,
                                             arg.col)

        msg = ("Argument {0} must fall on any of: "
               "{1}".format(arg.contents,
                            ", ".join(msg_parts)))
        return align, LinterFailure(msg, arg.line, replacement)

    return align, None


def func_args_aligned(contents, abstract_syntax_tree):
    """Check that function arguments are aligned.

    Function arguments must be aligned either to the same line or
    must fall on the same column as the last argument on the line of
    the function call. The only special case is for definitions, where we
    align after the second argument, (eg the first argument to the defined
    function or macro)
    """
    errors = []

    def _call_visitor(name, node):
        """Visit all function calls."""
        assert name == "FunctionCall"

        def _align_violations(node):
            """All alignment violations in node."""
            arguments_len = len(node.arguments)

            if arguments_len == 0:
                return

            # Check if this was a definition. If so, then we align after the
            # last argument and not after the first.
            is_definition = (_RE_IS_DEFINITION.match(node.name) is not None)

            if is_definition and arguments_len > 1:
                baseline_col = node.arguments[1].col
            else:
                baseline_col = node.arguments[0].col

            align = AlignmentInfo(col=node.arguments[0].col,
                                  line=node.arguments[0].line,
                                  line_has_kw=False)

            for index in range(0, len(node.arguments)):
                arg = node.arguments[index]
                # If the argument is on the same line as the function call,
                # then update align.col and make sure that the argument is one
                # space away from the last argument
                if align.line == arg.line:
                    yield _check_horizontal_space(node, index, contents)

                align, error = _check_alignment(arg,
                                                align,
                                                baseline_col,
                                                node.arguments[0].line,
                                                contents)

                yield error

        errors.extend([e for e in _align_violations(node) if e is not None])

    ast_visitor.recurse(abstract_syntax_tree,
                        function_call=ignore.visitor_depth(_call_visitor))
    return errors


def double_outer_quotes(contents, abstract_syntax_tree):
    """Check that all outer quotes are double quotes."""
    errors = []

    def _word_visitor(name, node):
        """Visit all arguments."""
        assert name == "Word"

        if node.type == WordType.String:
            if not _RE_DOUBLE_OUTER_QUOTES.match(node.contents):
                msg = "{0} must use double quotes".format(node.contents)
                replacement_word = "\"{0}\"".format(node.contents[1:-1])
                replacement = util.replace_word(contents[node.line - 1],
                                                node.col - 1,
                                                node.contents,
                                                replacement_word)
                errors.append(LinterFailure(msg, node.line, replacement))

    ast_visitor.recurse(abstract_syntax_tree,
                        word=ignore.visitor_depth(_word_visitor))

    return errors


# The mccabe checker complains here but that's because it contains lots
# of nested functions. There's not much we can do for that.
def _visit_with_flat_if_depth(abstract_syntax_tree,  # NOQA
                              function_visitor):
    """Visit function calls with IfBlock nodes flattened."""
    def _visit_header_body_block(node, depth):
        """Handle header/body like statements."""
        function_visitor(node.header, depth)

        for sub in node.body:
            _node_dispatch(sub, depth + 1)

        if getattr(node, "footer", None) is not None:
            function_visitor(node.footer, depth)

    def _visit_if_block(node, depth):
        """Handle if blocks."""
        substatements = ([node.if_statement] +
                         node.elseif_statements +
                         [node.else_statement])

        for substatement in substatements:
            if substatement:
                _visit_header_body_block(substatement, depth)

        function_visitor(node.footer, depth)

    def _visit_toplevel(node, depth):
        """Handle toplevel blocks."""
        assert depth == 0

        for statement in node.statements:
            _node_dispatch(statement, depth)

    indent_node_dispatch = [
        (_RE_HEADERBODY_LIKE, _visit_header_body_block),
        (_RE_IF_BLOCK, _visit_if_block),
        (_RE_FUNCTION_CALL, function_visitor),
        (_RE_TOPLEVEL, _visit_toplevel)
    ]

    def _node_dispatch(node, depth):
        """Dispatch per-node, handling depth like actual indent depth."""
        for regex, handler in indent_node_dispatch:
            if regex.match(node.__class__.__name__):
                handler(node, depth)

    _node_dispatch(abstract_syntax_tree, 0)


def calls_indented_correctly(contents, abstract_syntax_tree, **kwargs):
    """Check that all calls to functions are indented at the correct level."""
    errors = []

    try:
        indent = kwargs["indent"]
    except KeyError:
        return errors

    def _visit_function_call(node, depth):
        """Handle function calls."""
        col = node.col
        expected = 1 + (depth * indent)
        if node.col != expected:
            delta = expected - node.col
            msg = "Expected {0} to be on column {1}".format(node.name,
                                                            expected)
            replacement = util.replace_word(contents[node.line - 1],
                                            col - 1 + min(0, delta),
                                            " " * max(0, delta * -1),
                                            " " * max(0, delta))
            errors.append(LinterFailure(msg, node.line, replacement))

    _visit_with_flat_if_depth(abstract_syntax_tree, _visit_function_call)

    return errors



# /polysquarecmakelinter/check_access.py
#
# Linter checks for access rights
#
# See /LICENCE.md for Copyright information
"""Linter checks for access rights."""

from polysquarecmakelinter import find_all
from polysquarecmakelinter import find_variables_in_scopes

from polysquarecmakelinter.types import LinterFailure


def only_use_own_privates(abstract_syntax_tree):
    """Check that all private definitions used are defined here."""
    calls, defs = find_all.private_calls_and_definitions(abstract_syntax_tree)

    errors = []

    for call, info in calls.items():
        if call not in defs.keys():
            for line in info:
                msg = "Used external private definition {0}".format(call)
                errors.append(LinterFailure(msg, line))

    return errors


def _find_violating_priv_uses(variable, current_scope):
    """Generate a list of uses of private variables not in this module."""
    for use in find_all.variables_used_in_expr(variable.node):
        if use.startswith("_"):
            # Used a private, check if it was set
            private_var_was_set = False

            traverse_scope = current_scope
            while traverse_scope is not None:
                for set_var in traverse_scope.set_vars:
                    if (set_var.node.contents == use and
                            (set_var.node.line,
                             set_var.node.col) != variable):
                        private_var_was_set = True
                        break

                traverse_scope = traverse_scope.parent

            if not private_var_was_set:
                yield (use, variable.node.line)


def only_use_own_priv_vars(ast):
    """Check that all private variables used are defined here."""
    used_privs = []

    global_set_vars = find_variables_in_scopes.set_in_tree(ast)
    global_used_vars = find_variables_in_scopes.used_in_tree(ast)

    _, global_definitions = find_all.private_calls_and_definitions(ast)

    # The big assumption here is that the "scopes" structure in both
    # trees are the same
    def _scope_visitor(set_vars_scope, used_vars_scope):
        """Visit scope's set vars and used vars.

        If a var was private and used, but not set in this scope or any
        parents, then report an error
        """
        assert len(set_vars_scope.scopes) == len(used_vars_scope.scopes)

        for index in range(0, len(set_vars_scope.scopes)):
            _scope_visitor(set_vars_scope.scopes[index],
                           used_vars_scope.scopes[index])

        for variable in used_vars_scope.used_vars:
            used_privs.extend(list(_find_violating_priv_uses(variable,
                                                             set_vars_scope)))

    _scope_visitor(global_set_vars, global_used_vars)

    # Filter out definitions of private functions of the same name
    # as functions can be used as variables.
    used_privs = [up for up in used_privs if up[0] not in global_definitions]

    err_msg = "Referenced external private variable {0}"
    return [LinterFailure(err_msg.format(u[0]), u[1]) for u in used_privs]



# /test/test_correctness_warnings.py
#
# Test cases for correctness/* checks
#
# See /LICENCE.md for Copyright information
"""Test cases for correctness/* checks."""

from test.warnings_test_common import LinterFailure
from test.warnings_test_common import replacement
from test.warnings_test_common import run_linter_throw

from nose_parameterized import parameterized

from polysquarecmakelinter import check_correctness

from testtools import ExpectedException
from testtools import TestCase

# Generate some quote-only variable combinations
QUOTE_ONLY_TEST_VARIABLES = []
for variable in check_correctness.ALWAYS_QUOTE_VARIABLES_CONTAINING:
    QUOTE_ONLY_TEST_VARIABLES += [
        "${_" + variable.variable + "}",
        "${" + variable.variable + "}",
        "${A_" + variable.variable + "}",
        "${" + variable.variable + "}/Other",
        "Other/${" + variable.variable + "}"
    ]

NO_QUOTE_TEST_VARIABLES = []
for variable in check_correctness.ALWAYS_QUOTE_VARIABLES_CONTAINING:
    NO_QUOTE_TEST_VARIABLES += [
        "${A" + variable.variable + "}",
        "${" + variable.variable + "A}",
        "${" + variable.variable + "_}"
    ]


class TestQuoteVariablesWhichMayHaveSpaces(TestCase):
    """Test case for ensuring that certain variables are always quoted."""

    @parameterized.expand(QUOTE_ONLY_TEST_VARIABLES)
    def test_fail_deref_certain_vars(self, bad_deref):  # suppress(no-self-use)
        """Test correctness/quotes fails if dereffing some variable names."""
        with ExpectedException(LinterFailure):
            run_linter_throw("call ({0})".format(bad_deref),
                             whitelist=["correctness/quotes"])

    def test_fail_when_using_slashes(self):  # suppress(no-self-use)
        """Test correctness/quotes fails when using raw unquoted path."""
        with ExpectedException(LinterFailure):
            run_linter_throw("call (abc/def)\n",
                             whitelist=["correctness/quotes"])

    @parameterized.expand(NO_QUOTE_TEST_VARIABLES)
    def test_pass_deref_nonpath_var(self, deref):
        """Test correctness/quotes passes with modified path var name."""
        self.assertTrue(run_linter_throw("call ({0})".format(deref),
                                         whitelist=["correctness/quotes"]))

    def test_replace_when_using_slashes(self):
        """Test correctness/quotes replaces raw unquoted path with quotes."""
        def get_replacement():
            """Get the replacement for raw unquoted path."""
            run_linter_throw("call (abc/def)\n",
                             whitelist=["correctness/quotes"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, "call (\"abc/def\")\n"))

    @parameterized.expand(QUOTE_ONLY_TEST_VARIABLES)
    def test_replace_deref_certain_vars(self, bad_deref):
        """Test correctness/quotes replaces variable derefs with quotes."""
        def get_replacement():
            """Get the replacement for the dereferenced path."""
            run_linter_throw("call ({0})".format(bad_deref),
                             whitelist=["correctness/quotes"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, "call (\"{0}\")".format(bad_deref)))



# /test/test_acceptance.py
#
# Test cases for usage of polysquarecmakelinter.main() and command line
# parsing.
#
# See /LICENCE.md for Copyright information
"""Test cases for usage of polysquarecmakelinter.main()."""

import os

import tempfile

from polysquarecmakelinter import linter

from testtools import TestCase


def run_linter_main(filename, *args, **kwargs):
    """Run linter.main() (as an integration test)."""
    del args

    arguments = [filename]

    def _convert_kv_to_switches(key, value):
        """Convert a key-value pair to command-line switches."""
        append_args = ["--{0}".format(key).replace("_", "-")]

        type_dispatch = {
            bool: [],
            list: value,
            str: [value]
        }

        # We assume that the types in type_dispatch are the only types
        # we'll encounter, all others will throw an exception.
        append_args += type_dispatch[type(value)]
        return append_args

    for key, value in kwargs.items():
        arguments += _convert_kv_to_switches(key, value)

    return linter.main(arguments)


class TestLinterAcceptance(TestCase):
    """Acceptance tests for linter.main()."""

    def __init__(self, *args, **kwargs):
        """"Initialize class variables."""
        cls = TestLinterAcceptance
        super(cls, self).__init__(*args,  # suppress(R903)
                                  **kwargs)
        self._temporary_file = None

    def setUp(self):  # NOQA
        """Create a temporary file."""
        super(TestLinterAcceptance, self).setUp()
        self._temporary_file = tempfile.mkstemp()

    def tearDown(self):  # NOQA
        """Remove temporary file."""
        os.remove(self._temporary_file[1])
        super(TestLinterAcceptance, self).tearDown()

    def test_blacklist(self):
        """Check that blacklisting a test causes it not to run."""
        contents = "function_call()\n"

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 blacklist=["style/space_before_func"])
        self.assertEqual(result, 0)

    def test_whitelist_pass(self):
        """Check that whitelisting a test causes only it to run."""
        contents = "    FUNCTION_CALL ()\n"

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 whitelist=["style/space_before_func"])

        self.assertEqual(result, 0)

    def test_namespace_pass(self):
        """Check passing a test when using --namespace."""
        contents = "function (our_function)\nendfunction ()\n"

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 whitelist=["structure/namespace"],
                                 namespace="our")

        self.assertEqual(result, 0)

    def test_namespace_fail(self):
        """Check failing a test when using --namespace."""
        contents = "function (func)\nendfunction ()\n"

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 whitelist=["structure/namespace"],
                                 namespace="our")

        self.assertEqual(result, 1)

    def test_indent_pass(self):
        """Check passing a test when using --indent."""
        contents = "function (our_function)\n  call ()\nendfunction ()\n"

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 whitelist=["style/indent"],
                                 indent="2")

        self.assertEqual(result, 0)

    def test_indent_fail(self):
        """Check failing a test when using --indent."""
        contents = "function (our_function)\n  call ()\nendfunction ()\n"

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 whitelist=["style/indent"],
                                 indent="4")

        self.assertEqual(result, 1)

    def test_whitelist_fail(self):
        """Check that whitelisting a test causes only it to run."""
        contents = "    FUNCTION_CALL()\n"

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 whitelist=["style/space_before_func"])

        self.assertEqual(result, 1)

    def test_fix_what_you_can(self):
        """Check that --fix-what-you-can modifies file correctly."""
        contents = "function_call()\n"

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        run_linter_main(self._temporary_file[1],
                        whitelist=["style/space_before_func"],
                        fix_what_you_can=True)

        with open(self._temporary_file[1], "r") as processed_file:
            self.assertEqual("function_call ()\n", processed_file.read())



# /test/test_access_warnings.py
#
# Test cases for access/* checks
#
# See /LICENCE.md for Copyright information
"""Test cases for access/* checks."""

from test.warnings_test_common import DEFINITION_TYPES
from test.warnings_test_common import FUNCTIONS_SETTING_VARS
from test.warnings_test_common import LinterFailure
from test.warnings_test_common import format_with_command
from test.warnings_test_common import run_linter_throw

from nose_parameterized import param, parameterized

from testtools import (ExpectedException, TestCase)


class TestPrivateFunctionsUsedMustBeDefinedInThisModule(TestCase):
    """Test that private functions used are defined in this module."""

    @parameterized.expand(DEFINITION_TYPES)
    # suppress (no-self-use)
    def test_pass_used_own_priv_def(self, definition):
        """access/other_private passes if using own private definition."""
        script = ("{0} (_definition ARGUMENT)\n"
                  "end{0} ()\n"
                  "_definition (ARGUMENT)\n").format(definition)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["access/other_private"]))

    def test_pass_use_pub_def(self):
        """access/other_private passes if we use public def from other mod."""
        self.assertTrue(run_linter_throw("definition (ARGUMENT)\n",
                                         whitelist=["access/other_private"]))

    def test_fail_use_other_priv_def(self):  # suppress(no-self-use)
        """acess/other_private fails if we use private def from other mod."""
        with ExpectedException(LinterFailure):
            run_linter_throw("_definition (ARGUMENT)\n",
                             whitelist=["access/other_private"])

# suppress(unnecessary-lambda)
_PRIVATE_VAR_SET_FORMAT = format_with_command(lambda x: "_{}".format(x))


class TestNoUseOtherPrivateToplevelVars(TestCase):
    """Check that private variables not set by us are not used."""

    parameters = [param(m) for m in FUNCTIONS_SETTING_VARS]

    def test_pass_pub_var_used(self):
        """Check access/private_var passes if undefined public var is used."""
        script = "message (${VALUE})\n"
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["access/private_var"]))

    @parameterized.expand(parameters,
                          testcase_func_doc=_PRIVATE_VAR_SET_FORMAT)
    def test_pass_variable_used(self, matcher):
        """Check access/private_var passes when priv var set by {}."""
        # suppress(unnecessary-lambda,E731)
        xform = lambda x: "_{0}".format(x)
        private_var = matcher.find.generate(matcher.sub,
                                            lambda x: x, xform)
        script = ("{0} ({1})\n"
                  "message ({2})\n").format(matcher.cmd,
                                            private_var,
                                            "${_VALUE}")
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["access/private_var"]))

    def test_pass_priv_func_used(self):
        """Check access/private_var passes when using private func as var."""
        script = ("function (_private_function)\n"
                  "endfunction ()\n"
                  "call (_private_function)\n")
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["access/private_var"]))

    def test_pass_foreach_set(self):
        """Check access/private_var passes when private var set by foreach."""
        script = ("foreach (_LOOP_VAR ${LIST})\n"
                  "    message (STATUS \"${_LOOP_VAR}\")\n"
                  "endforeach ()\n")
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["access/private_var"]))

    def test_fail_set_outside_scope(self):  # suppress(no-self-use)
        """Check access/private_var fails when var set outside scope."""
        script = ("foreach (_LOOP_VAR ${LIST})\n"
                  "endforeach ()\n"
                  "message (STATUS \"${_LOOP_VAR}\")\n")
        with ExpectedException(LinterFailure):
            run_linter_throw(script, whitelist=["access/private_var"])

    def test_fail_priv_var_used(self):  # suppress(no-self-use)
        """Check access/private_var fails on undefined private var used."""
        script = "message (${_VALUE})\n"
        with ExpectedException(LinterFailure):
            run_linter_throw(script, whitelist=["access/private_var"])



# /test/test_style_warnings.py
#
# Test cases for style/* checks
#
# See /LICENCE.md for Copyright information
"""Test cases for style/* checks."""

from test.warnings_test_common import DEFINITION_TYPES
from test.warnings_test_common import FUNCTIONS_SETTING_VARS
from test.warnings_test_common import LinterFailure
from test.warnings_test_common import format_with_args
from test.warnings_test_common import format_with_command
from test.warnings_test_common import gen_source_line
from test.warnings_test_common import replacement
from test.warnings_test_common import run_linter_throw

from nose_parameterized import param, parameterized

from testtools import ExpectedException
from testtools import TestCase


class TestSpaceBeforeFunctionCallWarnings(TestCase):
    """Test case for a single space between a function call and name."""

    def test_lint_pass(self):
        """Check that style/space_before_func passes.

        Test passes where there is a single space before a function name
        and a call, like so:

        function_name ()
        """
        result = run_linter_throw("function_call ()\n",
                                  whitelist=["style/space_before_func"])
        self.assertTrue(result)

    def test_lint_pass_comment(self):
        """Check that style/space_before_func passes for commented calls.

        Test passes where there is no space before a function name
        and a call, where that line is commented like so:

        # function_name()
        """
        result = run_linter_throw("# function_call()\n",
                                  whitelist=["style/space_before_func"])
        self.assertTrue(result)

    def test_lint_pass_inside_quotes(self):
        """Check that style/space_before_func passes for quoted calls.

        Test passes where there is no space before a function name
        and a call, where that line is inside quotes

        "function_name()"
        """
        result = run_linter_throw("call (\"function_call()\")\n",
                                  whitelist=["style/space_before_func"])
        self.assertTrue(result)

    def test_lint_fail_nospace(self):   # suppress(no-self-use)
        """Check that style/space_before_func fails.

        Test fails where there is no space between a function name and a
        call, like so:

        function_name()
        """
        with ExpectedException(LinterFailure):
            run_linter_throw("function_call()\n",
                             whitelist=["style/space_before_func"])

    def test_lint_fail_excessive_space(self):  # suppress(no-self-use)
        """Check that style/space_before_func fails.

        Test fails where there is more than one space between a function name
        and a call, like so

        function_name ()
        """
        with ExpectedException(LinterFailure):
            run_linter_throw("function_call  ()\n",
                             whitelist=["style/space_before_func"])

    def test_replace_excess_one_space(self):
        """Check that the style/space_before_func replacement has one space."""
        def get_replacement():
            """Get replacement for function call with excessive whitespace."""
            run_linter_throw("function_call            ()\n",
                             whitelist=["style/space_before_func"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, "function_call ()\n"))

    def test_replace_nospace_one_space(self):
        """Check that the style/space_before_func replacement has one space."""
        def get_replacement():
            """Get replacement for function call with no whitespace."""
            run_linter_throw("function_call()\n",
                             whitelist=["style/space_before_func"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, "function_call ()\n"))


class TestFunctionsMustbeLowercaseOnly(TestCase):
    """Test case for functions and macros being lowercase."""

    def test_pass_lowercase_call(self):
        """style/lowercase passes when calling lowercase func."""
        result = run_linter_throw("lowercase_func (ARGUMENT)\n",
                                  whitelist=["style/lowercase_func"])
        self.assertTrue(result)

    def test_fail_uppercase_call(self):  # suppress(no-self-use)
        """style/lowercase fails when calling uppercase func."""
        with ExpectedException(LinterFailure):
            run_linter_throw("UPPERCASE_FUNC (ARGUMENT)\n",
                             whitelist=["style/lowercase_func"])

    def test_replace_uppercase_call(self):
        """style/lowercase replaces uppercase call with lowercase call."""
        func_name = "UPPERCASE_FUNC"
        error_line = "{0} (ARGUMENT)\n".format(func_name)
        replacement_line = "{0} (ARGUMENT)\n".format(func_name.lower())

        def get_replacement():
            """Replacement for all uppercase function call."""
            run_linter_throw(error_line,
                             whitelist=["style/lowercase_func"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, replacement_line))

    def test_pass_lowercase_func_def(self):
        """style/lowercase passes when defining lowercase func."""
        result = run_linter_throw("function (lowercase_func) endfunction ()\n",
                                  whitelist=["style/lowercase_func"])
        self.assertTrue(result)

    def test_fail_uppercase_func_def(self):  # suppress(no-self-use)
        """style/lowercase fails when defining uppercase func."""
        with ExpectedException(LinterFailure):
            run_linter_throw("function (UPPERCASE_FUNC) endfunction ()\n",
                             whitelist=["style/lowercase_func"])

    def test_replace_uppercase_func_def(self):
        """style/lowercase replaces uppercase call with lowercase call."""
        func_name = "UPPERCASE_FUNC"
        lower_name = func_name.lower()
        error = "function ({0}) endfunction ()\n".format(func_name)
        expected_repl = "function ({0}) endfunction ()\n".format(lower_name)

        def get_replacement():
            """Replace uppercase function call."""
            run_linter_throw(error,
                             whitelist=["style/lowercase_func"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, expected_repl))

    def test_pass_lowercase_macro_def(self):
        """style/lowercase passes when defining lowercase macro."""
        result = run_linter_throw("macro (lowercase_macro) endmacro ()\n",
                                  whitelist=["style/lowercase_func"])
        self.assertTrue(result)

    def test_fail_uppercase_macro(self):  # suppress(no-self-use)
        """style/lowercase fails when defining uppercase macro."""
        with ExpectedException(LinterFailure):
            run_linter_throw("macro (UPPERCASE_MACRO) endmacro ()\n",
                             whitelist=["style/lowercase_func"])

    def test_replace_uppercase_macro(self):
        """style/lowercase replaces uppercase definition with lowercase def."""
        macro_name = "UPPERCASE_MACRO"
        lower_name = macro_name.lower()
        error = "macro ({0}) endmacro ()\n".format(macro_name)
        expected_replacement = "macro ({0}) endmacro ()\n".format(lower_name)

        def get_replacement():
            """Replacement for uppercase macro."""
            run_linter_throw(error,
                             whitelist=["style/lowercase_func"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, expected_replacement))


class TestUppercaseDefinitionArguments(TestCase):
    """Check that all arguments to a definition are uppercase."""

    @parameterized.expand(DEFINITION_TYPES)
    def test_pass_no_args(self, defin):
        """Check style/uppercase_args passes where function has no args."""
        script = "{0} (definition_name)\nend{0} ()\n".format(defin)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["style/uppercase_args"]))

    @parameterized.expand(DEFINITION_TYPES)
    def test_pass_uppercase_args(self, defin):
        """Check style/uppercase_args passes where args are uppercase."""
        script = "{0} (definition_name UPPERCASE)\nend{0} ()\n".format(defin)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["style/uppercase_args"]))

    @parameterized.expand(DEFINITION_TYPES)
    def test_fail_lowercase_args(self, defin):  # suppress(no-self-use)
        """Check style/uppercase_args passes where args are lowercase."""
        script = "{0} (definition_name lowercase)\nend{0} ()\n".format(defin)
        with ExpectedException(LinterFailure):
            run_linter_throw(script, whitelist=["style/uppercase_args"])

    @parameterized.expand(DEFINITION_TYPES)
    def test_replace_with_upper(self, defin):
        """Check style/uppercase_args passes where args are lowercase."""
        script = "{0} (name lowercase)\nend{0} ()\n".format(defin)

        def get_replacement():
            """Replacement for lowercase argument."""
            run_linter_throw(script, whitelist=["style/uppercase_args"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, "{0} (name LOWERCASE)\n".format(defin)))


_FORMAT_WITH_DEREFFED_VAR = format_with_command(lambda x: "${" + x + "}")
_FORMAT_WITH_LOWERCASE_VAR = format_with_command(lambda x: x.lower())
_FORMAT_WITH_OTHER_QUOTES = format_with_command(other_xform=lambda x: ("\"" +
                                                                       x +
                                                                       "\""))
_FORMAT_QUOTES_AND_LOWER = format_with_command(var_xform=lambda x: x.lower(),
                                               other_xform=lambda x: ("\"" +
                                                                      x +
                                                                      "\""))


class TestUppercaseVariableNamesOnly(TestCase):
    """Test case for uppercase variable names only."""

    parameters = [param(m) for m in FUNCTIONS_SETTING_VARS]

    @parameterized.expand(parameters, testcase_func_doc=format_with_args(0))
    def test_pass_no_var_set(self, matcher):
        """Check that style/set_var_case passes with {0.cmd}.

        Where no variable is actually set, then there is no linter failure
        """
        # This will trip up matchers that match other arguments
        result = run_linter_throw("{0} ()\n".format(matcher.cmd),
                                  whitelist=["style/set_var_case"])
        self.assertTrue(result)

    @parameterized.expand(parameters,
                          testcase_func_doc=format_with_command())
    def test_pass_no_quotes(self, matcher):
        """Check that style/set_var_case passes with {}.

        Variables set by another CMake command should only be uppercase
        """
        result = run_linter_throw(gen_source_line(matcher),
                                  whitelist=["style/set_var_case"])
        self.assertTrue(result)

    @parameterized.expand(parameters,
                          testcase_func_doc=_FORMAT_WITH_DEREFFED_VAR)
    def test_pass_inside_deref(self, matcher):
        """Check that style/set_var_case passes when var in deref, like {}.

        Pass if variable is uppercase and inside of a deref, because variable
        dereferences are not sink variables.
        """
        xform = lambda x: "${" + x + "}"  # suppress(E731)
        result = run_linter_throw(gen_source_line(matcher,
                                                  match_transform=xform),
                                  whitelist=["style/set_var_case"])
        self.assertTrue(result)

    @parameterized.expand(parameters,
                          testcase_func_doc=_FORMAT_WITH_OTHER_QUOTES)
    def test_pass_other_quotes(self, matcher):
        """Check that style/set_var_case pass with other args quoted in {}."""
        quote = "\"{0}\""
        xform = lambda x: quote.format(x)  # suppress(unnecessary-lambda,E731)
        line = gen_source_line(matcher,
                               other_transform=xform)
        result = run_linter_throw(line,
                                  whitelist=["style/set_var_case"])
        self.assertTrue(result)

    @parameterized.expand(parameters,
                          testcase_func_doc=_FORMAT_WITH_LOWERCASE_VAR)
    def test_fail_no_quotes(self, matcher):  # suppress(no-self-use)
        """Check that style/set_var_case fails with {}, because lowercase."""
        line = gen_source_line(matcher,
                               match_transform=lambda x: x.lower())
        with ExpectedException(LinterFailure):
            run_linter_throw(line,
                             whitelist=["style/set_var_case"])

    @parameterized.expand(parameters,
                          testcase_func_doc=_FORMAT_QUOTES_AND_LOWER)
    def test_fail_other_quotes(self, matcher):  # suppress(no-self-use)
        """Check that style/set_var_case fails with other args quoted in {}."""
        quote = "\"{0}\""
        xform = lambda x: quote.format(x)  # suppress(unnecessary-lambda,E731)
        line = gen_source_line(matcher,
                               match_transform=lambda x: x.lower(),
                               other_transform=xform)
        with ExpectedException(LinterFailure):
            run_linter_throw(line,
                             whitelist=["style/set_var_case"])

    @parameterized.expand(parameters,
                          testcase_func_doc=_FORMAT_WITH_LOWERCASE_VAR)
    def test_replace_no_quotes(self, matcher):
        """Check that style/set_var_case replaces {} with uppercase var.

        Replacement should have uppercase matched argument
        """
        correct = gen_source_line(matcher)
        incorrect = gen_source_line(matcher,
                                    match_transform=lambda x: x.lower())

        def get_replacement():
            """Replacement for lowercase variable."""
            run_linter_throw(incorrect,
                             whitelist=["style/set_var_case"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, correct))


class TestFunctionArgumentsFallOnLine(TestCase):
    """Test alignment of function arguments."""

    def test_pass_args_on_same_line(self):
        """style/argument_align passes when args on same line."""
        self.assertTrue(run_linter_throw("call ($[ONE} TWO THREE \"FOUR\")\n",
                                         whitelist=["style/argument_align"]))

    def test_fail_args_unevenly_spaced(self):  # suppress(no-self-use)
        """style/argument_align fails if args on same line spaced unevenly."""
        with ExpectedException(LinterFailure):
            run_linter_throw("call (ONE   TWO)\n",
                             whitelist=["style/argument_align"])

    def test_suggest_even_spacing(self):
        """style/argument_align suggests even spacing on the same line."""
        def get_replacement():
            """Get replacement for unevenly spaced lines."""
            run_linter_throw("call (ONE   TWO)\n",
                             whitelist=["style/argument_align"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, "call (ONE TWO)\n"))

    def test_fail_args_not_aligned(self):  # suppress(no-self-use)
        """style/argument_align fails when args do not fall on baseline col."""
        with ExpectedException(LinterFailure):
            run_linter_throw("call (ONE\nTWO)\n",
                             whitelist=["style/argument_align"])

    def test_fail_args_dispersed(self):  # suppress(no-self-use)
        """style/argument_align fails if args on same line spaced unevenly."""
        with ExpectedException(LinterFailure):
            run_linter_throw("call (ONE\n"
                             "      ${TWO} \"THREE\"\n"
                             "      FOUR)\n",
                             whitelist=["style/argument_align"])

    def test_fail_bad_kw_align(self):  # suppress(no-self-use)
        """style/argument_align fails if args on same line spaced unevenly."""
        with ExpectedException(LinterFailure):
            run_linter_throw("call (ONE\n"
                             "      TWO THREE\n"
                             "        FOUR)\n",
                             whitelist=["style/argument_align"])

    def test_fail_inconsistent_align(self):  # suppress(no-self-use)
        """style/argument_align fails when args not aligned after first."""
        with ExpectedException(LinterFailure):
            run_linter_throw("call (${ONE} TWO\n"
                             "             THREE)\n",
                             whitelist=["style/argument_align"])

    # Over and under-indent
    @parameterized.expand([
        "     THREE)\n",
        "         THREE)\n"
    ])
    def test_suggest_baseline_align(self, third_line):
        """style/argument_align suggests alignment to the baseline."""
        def get_replacement():
            """Get replacement for unevenly spaced lines."""
            run_linter_throw("call (ONE\n"
                             "      TWO\n" +
                             third_line,
                             whitelist=["style/argument_align"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         # eg  call (ONE
                         (3, ("      THREE)\n")))

    def test_fail_align_func_name(self):  # suppress(no-self-use)
        """style/argument_align fails when args not aligned after second."""
        with ExpectedException(LinterFailure):
            run_linter_throw("function (ONE TWO\n"
                             "          THREE)\n"
                             "endfunction ()\n",
                             whitelist=["style/argument_align"])

    def test_fail_align_macro_name(self):  # suppress(no-self-use)
        """style/argument_align fails when args not aligned after second."""
        with ExpectedException(LinterFailure):
            run_linter_throw("macro (name TWO\n"
                             "       THREE)\n"
                             "endmacro ()\n",
                             whitelist=["style/argument_align"])

    def test_suggest_align_first_arg(self):
        """style/argument_align suggests alignment to function's first arg."""
        def get_replacement():
            """Get replacement for unevenly spaced lines."""
            run_linter_throw("function (name ONE\n"
                             "            TWO)\n"
                             "endfunction ()\n",
                             whitelist=["style/argument_align"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         # eg, function (name ONE
                         (2, ("               TWO)\n")))

    def test_pass_args_aligend(self):
        """style/argument_align passes when args aligned."""
        self.assertTrue(run_linter_throw("call (ONE\n"
                                         "      TWO)\n",
                                         whitelist=["style/argument_align"]))

    def test_pass_align_after(self):
        """style/argument_align passes when args aligned after first."""
        self.assertTrue(run_linter_throw("call (ONE TWO\n"
                                         "      THREE)\n",
                                         whitelist=["style/argument_align"]))

    def test_pass_args_after_keyword(self):
        """style/argument_align passes with args after keyword arg."""
        self.assertTrue(run_linter_throw("call (ONE\n"
                                         "      KEYWORD TWO\n"
                                         "      KEYWORD THREE)\n",
                                         whitelist=["style/argument_align"]))

    def test_pass_align_after_keyword(self):
        """style/argument_align passes with args after keyword arg."""
        self.assertTrue(run_linter_throw("call (ONE\n"
                                         "      KEYWORD TWO\n"
                                         "              THREE)\n",
                                         whitelist=["style/argument_align"]))

    nonvariable_keywords = [
        "${KEYWORD}",
        "\"KEYWORD\"",
        "KEYWORD/ARGUMENT\"",
        "1234"
    ]

    @parameterized.expand(nonvariable_keywords)
    def test_fail_if_kw_not_var_align(self, keyword):  # suppress(no-self-use)
        """style/argument_align fails when args not aligned after second."""
        kw_len = len(keyword)
        with ExpectedException(LinterFailure):
            run_linter_throw("call (ONE\n"
                             "      {0} ONE".format(keyword) +
                             "      " + " " * kw_len + " TWO)",
                             whitelist=["style/argument_align"])

    @parameterized.expand(nonvariable_keywords)
    def test_fail_if_kw_not_var_after(self, keyword):  # suppress(no-self-use)
        """style/argument_align fails when args not aligned after second."""
        with ExpectedException(LinterFailure):
            run_linter_throw("call (ONE\n"
                             "      {0} ONE)\n".format(keyword),
                             whitelist=["style/argument_align"])

    def test_pass_align_after_func(self):
        """style/argument_align passes when args aligned after second."""
        self.assertTrue(run_linter_throw("function (name TWO\n"
                                         "               THREE)\n"
                                         "endfunction ()\n",
                                         whitelist=["style/argument_align"]))

    def test_pass_align_after_macro(self):
        """style/argument_align passes when args aligned after second."""
        self.assertTrue(run_linter_throw("macro (name TWO\n"
                                         "            THREE)\n"
                                         "endmacro ()\n",
                                         whitelist=["style/argument_align"]))

    def test_pass_dispersed_if_cond(self):
        """style/argument_align passes when arguments to if are dispersed."""
        self.assertTrue(run_linter_throw("if (CONDITION AND OTHER_COND OR\n"
                                         "    FINAL_CONDITION AND NOT COND)\n"
                                         "endif ()",
                                         whitelist=["unused/private_var"]))


class TestSingleQuoteUsage(TestCase):
    """Test that we are only allowed to use double quotes for strings."""

    def test_pass_use_double_quotes(self):
        """Check style/doublequotes passes when strings use double quotes."""
        self.assertTrue(run_linter_throw("call (\"ARGUMENT\")\n",
                                         whitelist=["style/doublequotes"]))

    def test_pass_sigle_in_double(self):
        """Check style/doublequotes passes if strings use internal single."""
        self.assertTrue(run_linter_throw("call (\"\'ARGUMENT\'\")\n",
                                         whitelist=["style/doublequotes"]))

    def test_fail_use_single_quotes(self):  # suppress(no-self-use)
        """Check style/doublequotes fails when strings use single quotes."""
        with ExpectedException(LinterFailure):
            run_linter_throw("call (\'ARGUMENT\')\n",
                             whitelist=["style/doublequotes"])

    def test_replace_single_with_double(self):
        """Check style/doublequotes replaces single quote use with double."""
        def get_replacement():
            """Replacement for single outer quotes."""
            run_linter_throw("call (\'ARGUMENT\')\n",
                             whitelist=["style/doublequotes"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, "call (\"ARGUMENT\")\n"))

    def test_replace_only_outerquotes(self):
        """Check style/doublequotes only replaces outer quotes."""
        def get_replacement():
            """Replacement for single outer quote."""
            run_linter_throw("call (\'ARG \\'U\\' MENT\')\n",
                             whitelist=["style/doublequotes"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, "call (\"ARG \\'U\\' MENT\")\n"))


HEADER_BODY_STRUCTURES = [
    "function",
    "macro",
    "while",
    "foreach",
    "if"
]


class TestIndentation(TestCase):
    """Test indentation checks."""

    def test_pass_no_indent_spec(self):
        """style/indent passes when no indentation is specified."""
        self.assertTrue(run_linter_throw("function_call ()\n",
                                         whitelist=["style/indent"]))

    def test_pass_top_call_noindent(self):
        """style/indent passes with zero indents for toplevel calls."""
        self.assertTrue(run_linter_throw("function_call ()\n",
                                         whitelist=["style/indent"],
                                         indent=1))

    def test_pass_top_def_noindent(self):
        """style/indent passes with zero indents for toplevel definitions."""
        self.assertTrue(run_linter_throw("function (f ARG)\nendfunction()\n",
                                         whitelist=["style/indent"],
                                         indent=1))

    def test_pass_call_one_indent(self):
        """style/indent passes with one indent for nested calls."""
        script = "function (f ARG)\n call (ARG)\nendfunction ()"
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["style/indent"],
                                         indent=1))

    def test_pass_if_body_one_indent(self):
        """style/indent passes with one indent for if body."""
        script = "if (COND)\n call (ARG)\nendif ()"
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["style/indent"],
                                         indent=1))

    def test_pass_nest_if_indent(self):
        """style/indent passes with one indent for if body."""
        script = "if (COND)\n if (OTHER)\n  call (ARG)\n endif ()\nendif ()"
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["style/indent"],
                                         indent=1))

    def test_fail_one_indent_top_call(self):  # suppress(no-self-use)
        """style/indent fails with one indent for toplevel calls."""
        with ExpectedException(LinterFailure):
            run_linter_throw(" function_call ()\n",
                             whitelist=["style/indent"],
                             indent=1)

    def test_fail_one_indent_toplevel(self):  # suppress(no-self-use)
        """style/indent fails with one indent for toplevel defs."""
        with ExpectedException(LinterFailure):
            run_linter_throw(" function (definition ARG)\n endfunction ()",
                             whitelist=["style/indent"],
                             indent=1)

    @parameterized.expand(HEADER_BODY_STRUCTURES)
    def test_fail_bad_term_indent(self, structure):  # suppress(no-self-use)
        """style/indent fails with one indent terminator."""
        with ExpectedException(LinterFailure):
            run_linter_throw("{0} ()\n end{0} ()".format(structure),
                             whitelist=["style/indent"],
                             indent=1)

    @parameterized.expand([
        "else",
        "elseif"
    ])  # suppress(no-self-use)
    def test_fail_mismatch_if_alt(self, alt):
        """style/indent fails when else, elseif has mismatched indent."""
        with ExpectedException(LinterFailure):
            script = "if (COND)\n {0} (COND)\nendif ()"
            run_linter_throw(script.format(alt),
                             whitelist=["style/indent"],
                             indent=1)

    def test_fail_noindent_nested_call(self):  # suppress(no-self-use)
        """style/indent fails with zero indents for a nested call."""
        with ExpectedException(LinterFailure):
            script = "function (f ARG)\ncall (ARG)\nendfunction ()"
            run_linter_throw(script, whitelist=["style/indent"], indent=1)

    def test_suggest_more_indent(self):
        """style/indent suggests more indentation where required."""
        script = "function (f ARG)\ncall (ARG)\nendfunction ()"

        def get_replacement():
            """Replacement for lack of indent."""
            run_linter_throw(script, whitelist=["style/indent"], indent=1)

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (2, " call (ARG)\n"))

    def test_suggest_less_indent(self):
        """style/indent suggests less indentation where required."""
        script = "function (f ARG)\n call (ARG)\n endfunction ()\n"

        def get_replacement():
            """Replacement for too much indent."""
            run_linter_throw(script, whitelist=["style/indent"], indent=1)

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (3, "endfunction ()\n"))



# /polysquarecmakelinter/find_all.py
#
# Details to find certain occurrences of nodes
#
# See /LICENCE.md for Copyright information
"""Detail to find certain occurrences of nodes."""

import re

from cmakeast import ast_visitor

from polysquarecmakelinter import find_set_variables
from polysquarecmakelinter import find_variables_in_scopes
from polysquarecmakelinter import util

_RE_VARIABLE_USE = re.compile(r"(?<![^\${])[0-9A-Za-z_]+(?![^]}])")


def _append_line_occurence(tracker, name, line):
    """Append line to name entry in tracker."""
    if name not in tracker.keys():
        tracker[name] = []

    tracker[name].append(line)


def calls(abstract_syntax_tree, track_call):
    """Return a dict of calls mapped to where they occurred."""
    call_lines = {}

    def _call_handler(name, node, depth):
        """Visit all calls in this module."""
        assert name == "FunctionCall"

        del depth

        if track_call(node):
            _append_line_occurence(call_lines, node.name, node.line)

    ast_visitor.recurse(abstract_syntax_tree,
                        function_call=_call_handler)

    return call_lines


def definitions(abstract_syntax_tree, track_definition):
    """Return a dict of definitions mapped to where they occurred."""
    definition_lines = {}

    def _definition_handler(name, node, depth):
        """Visit all definitions."""
        assert name == "FunctionDefinition" or name == "MacroDefinition"
        assert len(node.header.arguments) > 0

        del depth

        if track_definition(node):
            _append_line_occurence(definition_lines,
                                   node.header.arguments[0].contents,
                                   node.line)

    ast_visitor.recurse(abstract_syntax_tree,
                        function_def=_definition_handler,
                        macro_def=_definition_handler)

    return definition_lines


def private_calls_and_definitions(abstract_syntax_tree):
    """Return a tuple of all private calls and definitions."""
    def _definition_is_private(node):
        """Check if a definition is private."""
        return node.header.arguments[0].contents.startswith("_")

    private_calls = calls(abstract_syntax_tree,
                          lambda x: x.name.startswith("_"))
    private_defs = definitions(abstract_syntax_tree,
                               _definition_is_private)

    return (private_calls, private_defs)


def _append_to_set_variables(name, node, set_variables):
    """Append to a dict of set variables."""
    if name not in set_variables.keys():
        set_variables[name] = []

    set_variables[name].append((node.line, node.col))


def toplevel_set_private_vars(abstract_syntax_tree):
    """Find all toplevel private variables.

    Returns a dict of variable names to places where such variables were set
    """
    set_variables = {}

    for statement in abstract_syntax_tree.statements:

        # We only want bare function calls, not definitions or the like
        if statement.__class__.__name__ != "FunctionCall":
            continue

        # Scan the statement for any variables set and append
        set_variable = find_set_variables.by_function_call(statement)

        if set_variable is not None:
            if (set_variable and
                    util.is_word_sink_variable(set_variable.type) and
                    set_variable.contents.startswith("_")):
                _append_to_set_variables(set_variable.contents,
                                         set_variable,
                                         set_variables)

    return set_variables


def variables_used_in_expr(word_node):
    """Return a set of variable names "used" by a node."""
    assert word_node.__class__.__name__ == "Word"

    used_variables_set = {}
    uses = _RE_VARIABLE_USE.findall(word_node.contents)

    for name in uses:
        if name not in used_variables_set.keys():
            used_variables_set[name] = []

        used_variables_set[name].append((word_node.line, word_node.col))

    return used_variables_set


def variables_used_matching(abstract_syntax_tree,
                            node_matcher,
                            name_matcher):
    """Return a set of variable names used whose nodes satisfy matchers."""
    variables_used = {}

    global_scope = find_variables_in_scopes.used_in_tree(abstract_syntax_tree)

    def _visit_scope(scope):
        """Visit a scope."""
        for subscope in scope.scopes:
            _visit_scope(subscope)

        for word, _ in scope.used_vars:

            if not node_matcher(word):
                continue

            variable_uses = _RE_VARIABLE_USE.findall(word.contents)

            for match in variable_uses:
                if not name_matcher(match):
                    continue

                _append_to_set_variables(match, word, variables_used)

    _visit_scope(global_scope)

    return variables_used



# /test/warnings_test_common.py
#
# Some common functionality for all of the warnings tests
#
# See /LICENCE.md for Copyright information
"""Some common functionality for all of the warnings tests."""

from polysquarecmakelinter import find_set_variables
from polysquarecmakelinter import linter

from testtools.matchers import Equals as TTEqMatcher


def _arg_no_mixin(num):
    """Return FindArgNoMixin for num."""
    class FindArgNoMixin(object):  # suppress(too-few-public-methods)
        """Mixin that provides generate() for finding args by number."""

        def generate(self,  # suppress(no-self-use)
                     sub,
                     other_transform,
                     match_transform):
            """Generate argument list of num size."""
            # A cheap hack to ensure that we generate a few extra in the
            # list GET case
            gen_num = abs(num - 1 if sub is not None else num)
            args_before = (" ".join([other_transform("ARGUMENT")] *
                                    gen_num))
            match_arg = match_transform("VALUE")
            subcommand = sub + " " if sub is not None else ""
            return subcommand + " ".join([args_before, match_arg])

    return FindArgNoMixin


def _after_arg_mixin(argument):
    """Return FindAfterArgMixin for argument."""
    class FindAfterArgMixin(object):  # suppress(too-few-public-methods)
        """Mixin that provides generate() for finding args after args."""

        def generate(self,  # suppress(no-self-use)
                     sub,
                     other_transform,
                     match_transform):
            """Generate argument list of ARGUMENT__ {argument} VALUE."""
            args_before = other_transform("ARGUMENT__ {0}".format(argument))
            match_arg = match_transform("VALUE")
            subcommand = sub + " " if sub is not None else ""
            return subcommand + " ".join([args_before, match_arg])

    return FindAfterArgMixin

FUNCTIONS_SETTING_VARS = find_set_variables.all_functions(_arg_no_mixin,
                                                          _after_arg_mixin)


# Pychecker complains about the Equals matcher failing to override comparator
# so do that here
class Equals(TTEqMatcher):  # suppress(R0903)
    """Matcher which tests equality."""

    def __init__(self, matchee):
        """Forward matchee to parent class."""
        super(Equals, self).__init__(matchee)

    def comparator(self, expected, other):
        """Check that expected == other."""
        return other == expected


class LinterFailure(Exception):
    """Exception raised when the linter reports a message."""

    def __init__(self, message, repl):
        """Initialize exception with message and replacement."""
        super(LinterFailure, self).__init__()
        self.message = message
        self.replacement = repl

    def __str__(self):
        """Return exception as string."""
        return str("{0}".format(self.message))


def run_linter_throw(contents, whitelist=None, blacklist=None, **kwargs):
    """Run linter.lint and throws if it reports a message."""
    errors = linter.lint(contents,
                         whitelist=whitelist,
                         blacklist=blacklist,
                         **kwargs)

    if len(errors):
        raise LinterFailure("{0} [{1}]".format(errors[0][1].line,
                                               errors[0][0]),
                            (errors[0][1].line, errors[0][1].replacement))

    return True


def replacement(exception):
    """Return replacement for a LinterFailure, else asserts."""
    assert exception.__class__.__name__ == "LinterFailure"
    return exception.replacement


def gen_source_line(matcher, match_transform=None, other_transform=None):
    """Generate a source line for use with tests, no quotes used."""
    m_xform = match_transform if match_transform is not None else lambda x: x
    o_xform = other_transform if other_transform is not None else lambda x: x
    finder = matcher.find
    line = "{0} (".format(matcher.cmd)
    line += finder.generate(matcher.sub, o_xform, m_xform)
    line += ")\n"
    return line


def format_with_command(var_xform=None,
                        other_xform=None):
    """Return function formatting docstring with 'set variable' tuple.

    Specify var_xform to transform the name of the variable
    which is set by the command.
    """
    def formatter(func, _, params):
        """Return formatted docstring with command setting variable."""
        source_line = gen_source_line(params.args[0],
                                      match_transform=var_xform,
                                      other_transform=other_xform)
        return func.__doc__.format(source_line)

    return formatter


def format_with_args(*args):
    """Return a function that formats a docstring."""
    def formatter(func, _, params):
        """Return formatted docstring with argument numbers in args."""
        pa = params.args
        format_args = [pa[i] for i in range(0, len(pa)) if i in args]

        return func.__doc__.format(*format_args)

    return formatter

DEFINITION_TYPES = [
    "function",
    "macro"
]



# /test/test_structure_warnings.py
#
# Test cases for structure/* checks
#
# See /LICENCE.md for Copyright information
"""Test cases for correctness/* checks."""

from test.warnings_test_common import DEFINITION_TYPES
from test.warnings_test_common import LinterFailure
from test.warnings_test_common import replacement
from test.warnings_test_common import run_linter_throw

from nose_parameterized import parameterized

from testtools import ExpectedException
from testtools import TestCase


class TestFunctionsMustBeNamespaces(TestCase):
    """Test that all functions must be namespaced."""

    @parameterized.expand(DEFINITION_TYPES)
    def test_pass_pub_func_namespaced(self, definition):
        """structure/namespace passes when public functions namespaced."""
        script = "{0} (our_call ARGUMENT)\nend{0} ()".format(definition)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["structure/namespace"],
                                         namespace="our"))

    @parameterized.expand(DEFINITION_TYPES)
    def test_pass_priv_func_namespaced(self, definition):
        """structure/namespace passes when private functions namespaced."""
        script = "{0} (_our_call ARGUMENT)\nend{0} ()".format(definition)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["structure/namespace"],
                                         namespace="our"))

    @parameterized.expand(DEFINITION_TYPES)
    def test_pass_no_namespace_passed(self, definition):
        """structure/namespace passes when no namespace specified."""
        script = "{0} (call ARGUMENT)\nend{0} ()".format(definition)
        self.assertTrue(run_linter_throw(script,
                                         whitelist=["structure/namespace"]))

    @parameterized.expand(DEFINITION_TYPES)
    def test_fail_outside_namespace(self, definition):  # suppress(no-self-use)
        """structure/namespace fails when definition outside namespace."""
        script = "{0} (not_our_call ARGUMENT)\nend{0} ()".format(definition)
        with ExpectedException(LinterFailure):
            run_linter_throw(script, namespace="our",
                             whitelist=["structure/namespace"])

    @parameterized.expand(DEFINITION_TYPES)
    # suppress(no-self-use)
    def test_fail_priv_out_namespace(self, definition):
        """structure/namespace fails when priv definition outside namespace."""
        script = "{0} (_not_our_call ARGUMENT)\nend{0} ()".format(definition)
        with ExpectedException(LinterFailure):
            run_linter_throw(script, namespace="our",
                             whitelist=["structure/namespace"])

    @parameterized.expand(DEFINITION_TYPES)
    def test_prepend_with_namespace(self, definition):
        """structure/namespace prepends namespace as replacement."""
        name = "not_our_call"
        script = "{0} ({1} ARGUMENT)\nend{0} ()".format(definition, name)

        def get_replacement():
            """Get replacement for un-namespaced definition."""
            run_linter_throw(script, namespace="our",
                             whitelist=["structure/namespace"])

        namespaced_name = "our_{0}".format(name)
        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, "{0} ({1} ARGUMENT)\n".format(definition,
                                                           namespaced_name)))

    @parameterized.expand(DEFINITION_TYPES)
    def test_priv_prepend_namespace(self, definition):
        """structure/namespace prepends namespace after _ as replacement."""
        name = "_not_our_call"
        script = "{0} ({1} ARGUMENT)\nend{0} ()".format(definition, name)

        def get_replacement():
            """Get replacement for un-namespaced private definition."""
            run_linter_throw(script, namespace="our",
                             whitelist=["structure/namespace"])

        namespaced_name = "_our_{0}".format(name)
        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, "{0} ({1} ARGUMENT)\n".format(definition,
                                                           namespaced_name)))



# /polysquarecmakelinter/linter.py
#
# Entry point for linter.
#
# See /LICENCE.md for Copyright information
"""Main module for linter."""

import argparse

import os

import re

import sys

from collections import OrderedDict

from cmakeast import ast

from jobstamps import jobstamp

from polysquarecmakelinter import check_access as access
from polysquarecmakelinter import check_correctness as correct
from polysquarecmakelinter import check_structure as structure
from polysquarecmakelinter import check_style as style
from polysquarecmakelinter import check_unused as unused
from polysquarecmakelinter import ignore

_RE_NOLINT = re.compile(r"^.*#\s+NOLINT:")


def should_ignore(line, warning):
    """Specify whether or not to ignore warnings on this line."""
    match = _RE_NOLINT.search(line)
    if match:
        try:
            # Special case, "*" means "all errors"
            if line[match.end():][0] == "*":
                return True
            elif line[match.end():match.end() + len(warning)] == warning:
                return True
        except IndexError:  # suppress(pointless-except)
            pass

    return False


LINTER_FUNCTIONS = {
    "structure/namespace": structure.definitions_namespaced,
    "style/space_before_func": ignore.check_kwargs(style.space_before_call),
    "style/set_var_case": ignore.check_kwargs(style.set_variables_capitalized),
    "style/uppercase_args": ignore.check_kwargs(style.uppercase_arguments),
    "style/lowercase_func": ignore.check_kwargs(style.lowercase_functions),
    "style/argument_align": ignore.check_kwargs(style.func_args_aligned),
    "style/doublequotes": ignore.check_kwargs(style.double_outer_quotes),
    "style/indent": style.calls_indented_correctly,
    "correctness/quotes": ignore.check_kwargs(correct.path_variables_quoted),
    "unused/private": ignore.all_but_ast(unused.private_definitions_used),
    "unused/var_in_func": ignore.all_but_ast(unused.vars_in_func_used),
    "unused/private_var": ignore.all_but_ast(unused.private_vars_at_toplevel),
    "access/other_private": ignore.all_but_ast(access.only_use_own_privates),
    "access/private_var": ignore.all_but_ast(access.only_use_own_priv_vars)
}


def lint(contents,
         whitelist=None,
         blacklist=None,
         **kwargs):
    r"""Actually lints some file contents.

    Contents should be a raw string with \n. whitelist is a list of checks
    to only perform, blacklist is list of checks to never perform.
    """
    abstract_syntax_tree = ast.parse(contents)
    contents_lines = contents.splitlines(True)
    linter_functions = LINTER_FUNCTIONS

    def _keyvalue_pair_if(dictionary, condition):
        """Return a key-value pair in dictionary if condition matched."""
        return {
            k: v for (k, v) in dictionary.items() if condition(k)
        }

    def _check_list(check_list, cond):
        """Return filter function for cond."""
        def _check_against_list(key):
            """Return true if list exists and condition passes."""
            return cond(check_list, key) if check_list is not None else True

        return _check_against_list

    linter_functions = _keyvalue_pair_if(linter_functions,
                                         _check_list(whitelist,
                                                     lambda l, k: k in l))
    linter_functions = _keyvalue_pair_if(linter_functions,
                                         _check_list(blacklist,
                                                     lambda l, k: k not in l))

    linter_errors = []
    for (code, function) in linter_functions.items():
        errors = function(contents_lines,
                          abstract_syntax_tree,
                          **kwargs)
        for error in errors:
            linter_errors.append((code, error))

    return linter_errors


# suppress(too-few-public-methods)
class ShowAvailableChecksAction(argparse.Action):
    """If --checks is encountered, just show available checks and exit."""

    def __call__(self, parser, namespace, values, option_string=None):
        """"Execute action."""
        del namespace
        del parser
        del values

        if option_string == "--checks":
            sys.stdout.write("Available option are:\n")
            for item in LINTER_FUNCTIONS.keys():
                sys.stdout.write(" * {0}\n".format(item))

            sys.exit(0)


def _parse_arguments(arguments=None):
    """Return a parser context result."""
    parser = argparse.ArgumentParser(description="Lint for Polysquare "
                                     "style guide")
    parser.add_argument("--checks",
                        nargs=0,
                        action=ShowAvailableChecksAction,
                        help="""list available checks""")
    parser.add_argument("files",
                        nargs="*",
                        metavar=("FILE"),
                        help="read FILE",
                        type=str)
    parser.add_argument("--whitelist",
                        nargs="*",
                        help="""list of checks that should only be run""",
                        default=None)
    parser.add_argument("--blacklist",
                        nargs="*",
                        help="""list of checks that should never be run""",
                        default=None)
    parser.add_argument("--indent",
                        nargs=1,
                        type=int,
                        help="""indent level""",
                        default=None)
    parser.add_argument("--namespace",
                        nargs=1,
                        type=str,
                        help="""namespace for functions""",
                        default=None)
    parser.add_argument("--fix-what-you-can",
                        action="store_true",
                        help="""automatically fix errors""")
    parser.add_argument("--stamp-directory",
                        type=str,
                        help="""directory to store cached results""")

    return parser.parse_args(arguments)


def _report_lint_error(error, file_path):
    """Report a linter error."""
    line = error[1].line
    code = error[0]
    description = error[1].description
    sys.stderr.write("{0}:{1} [{2}] {3}".format(file_path,
                                                line,
                                                code,
                                                description))


def _apply_replacement(error, found_file, file_lines):
    """Apply a single replacement."""
    fixed_lines = file_lines
    fixed_lines[error[1].line - 1] = error[1].replacement
    concatenated_fixed_lines = "".join(fixed_lines)

    # Only fix one error at a time
    found_file.seek(0)
    found_file.write(concatenated_fixed_lines)
    found_file.truncate()


def _jobstamps_kwargs(file_path, cache_output_directory):
    """Create keyword arguments to pass to jobstamps."""
    return {
        "jobstamps_dependencies": [file_path],
        "jobstamps_cache_output_directory": cache_output_directory
    }


def _sorted_if_exists(list_object):
    """Return sorted list if it exists."""
    return sorted(list_object) if list_object else None


def main(arguments=None):
    """Entry point for the linter."""
    result = _parse_arguments(arguments)

    num_errors = 0
    for found_file_name in result.files:
        with open(found_file_name, "r+") as found_file:
            file_path = os.path.abspath(found_file_name)
            file_contents = found_file.read()
            file_lines = file_contents.splitlines(True)
            try:
                kwargs = OrderedDict()
                if result.namespace is not None:
                    kwargs["namespace"] = result.namespace[0]

                if result.indent is not None:
                    kwargs["indent"] = result.indent[0]

                kwargs.update(_jobstamps_kwargs(file_path,
                                                result.stamp_directory))

                errors = jobstamp.run(lint,
                                      file_contents,  # suppress(star-args)
                                      _sorted_if_exists(result.whitelist),
                                      _sorted_if_exists(result.blacklist),
                                      **kwargs)
            except RuntimeError as err:
                msg = "RuntimeError in processing {0} - {1}".format(file_path,
                                                                    str(err))
                raise RuntimeError(msg)

            for error in errors:
                if not should_ignore(file_lines[error[1].line - 1], error[0]):
                    _report_lint_error(error, file_path)
                    if (result.fix_what_you_can and
                            error[1].replacement is not None):
                        _apply_replacement(error, found_file, file_lines)
                        sys.stderr.write(" ... FIXED\n")
                        break

                    sys.stderr.write("\n")

                    num_errors += 1

    return num_errors


if __name__ == "__main__":
    main()



# /polysquarecmakelinter/check_correctness.py
#
# Linter checks for potentially buggy behavior
#
# See /LICENCE.md for Copyright information
"""Linter checks for potentially buggy behavior."""

import re

from collections import namedtuple

from cmakeast import ast_visitor

from cmakeast.ast import WordType

from polysquarecmakelinter import util

from polysquarecmakelinter.types import LinterFailure

_RE_PATH_SLASH = re.compile(r"[\\/]")

_AlwaysQuote = namedtuple("_AlwaysQuoteT", "variable regex")


def _always_quote_if_at_end(variable):
    """Return an _AlwaysQuote with regex matching variable at end."""
    regex = r"\${.*(?<=[_{])" + variable + "}"
    return _AlwaysQuote(variable, re.compile(regex))

ALWAYS_QUOTE_VARIABLES_CONTAINING = [
    _always_quote_if_at_end("PATH"),
    _always_quote_if_at_end("FILE"),
    _always_quote_if_at_end("SOURCE"),
    _always_quote_if_at_end("HEADER"),
    _always_quote_if_at_end("DIRECTORY"),
    _always_quote_if_at_end("EXECUTABLE"),
    _always_quote_if_at_end("DIR"),
]

_ALWAYS_QUOTE_MATCHERS_INT = (ALWAYS_QUOTE_VARIABLES_CONTAINING +
                              [_AlwaysQuote("CMAKE_COMMAND",
                                            re.compile(r"\${CMAKE_COMMAND}"))])


def path_variables_quoted(contents, abstract_syntax_tree):
    """Check that each variable mutated is capitalized."""
    errors = []

    def _word_visitor(name, node, depth):
        """Visit all words."""
        assert name == "Word"

        del depth

        def _generate_error(node):
            """Generate an error and replacement for node in violation."""
            msg = "Path {0} must be quoted".format(node.contents)
            line_index = node.line - 1
            col_index = node.col - 1
            quoted = "\"{0}\""
            replacement = util.replace_word(contents[line_index],
                                            col_index,
                                            node.contents,
                                            quoted.format(node.contents))
            return LinterFailure(msg, node.line, replacement)

        if util.is_word_maybe_path(node.type):
            # CompoundLiterals definitely cannot have slashes
            if node.type == WordType.CompoundLiteral:
                if _RE_PATH_SLASH.search(node.contents):
                    errors.append(_generate_error(node))
                    return

            for always_quote_word in _ALWAYS_QUOTE_MATCHERS_INT:
                result = always_quote_word.regex.search(node.contents)
                if result is not None:
                    errors.append(_generate_error(node))
                    return

    ast_visitor.recurse(abstract_syntax_tree, word=_word_visitor)

    return errors



# /setup.py
#
# Installation and setup script for polysquare-cmake-linter
#
# See /LICENCE.md for Copyright information
"""Installation and setup script for polysquare-cmake-linter."""

from setuptools import find_packages
from setuptools import setup

setup(name="polysquare-cmake-linter",
      version="0.0.16",
      description="""Polysquare CMake Linter""",
      long_description_markdown_filename="README.md",
      author="Sam Spilsbury",
      author_email="smspillaz@gmail.com",
      url="http://github.com/polysquare/polysquare-cmake-linter",
      classifiers=["Development Status :: 3 - Alpha",
                   "Programming Language :: Python :: 2",
                   "Programming Language :: Python :: 2.7",
                   "Programming Language :: Python :: 3",
                   "Programming Language :: Python :: 3.1",
                   "Programming Language :: Python :: 3.2",
                   "Programming Language :: Python :: 3.3",
                   "Programming Language :: Python :: 3.4",
                   "Intended Audience :: Developers",
                   "Topic :: Software Development :: Build Tools",
                   "License :: OSI Approved :: MIT License"],
      license="MIT",
      keywords="development linters",
      packages=find_packages(exclude=["tests"]),
      install_requires=["cmakeast>=0.0.7",
                        "jobstamps>=0.0.16"],
      extras_require={
          "upload": ["setuptools-markdown"]
      },
      entry_points={
          "console_scripts": [
              "polysquare-cmake-linter=polysquarecmakelinter.linter:main"
          ]
      },
      test_suite="nose.collector",
      zip_safe=True,
      include_package_data=True)



# /polysquarecmakelinter/types.py
#
# Some types that are shared amongst linters and checks
#
# See /LICENCE.md for Copyright information
"""Some types that are shared amongst linters and checks."""

from collections import namedtuple


# Subclass instead of assignment so that we have override __new__
# and provide a default value
#
# suppress(too-few-public-methods)
class LinterFailure(namedtuple("LinterFailure",
                               "description line replacement")):
    """An immutable type representing a linter failure."""

    def __new__(cls, description, line, replacement=None):
        """Factory function."""
        return super(LinterFailure, cls).__new__(cls,
                                                 description,
                                                 line,
                                                 replacement)
